from trosnoth.model.universe_base import NeutralTeamId
from trosnoth.model.voteArbiter import VoteArbiter
from trosnoth.utils.utils import Static
from trosnoth.utils import math
from trosnoth.messages import TrosballTagMsg, GameOverMsg
import random

gameStatesById = {}

def registerGameState(gameStateClass):
    '''
    Allows a game state to be used across the network
    '''
    gameStatesById[gameStateClass.id] = gameStateClass
    return gameStateClass

def preferredTeamOtherwiseSmallest(preferredTeamId, universe):
    if preferredTeamId != NeutralTeamId:
        return preferredTeamId
    else:
        playerCounts = universe.getTeamPlayerCounts()

        minCount = len(universe.players) + 1
        minTeams = []
        for team in universe.teams:
            count = playerCounts.get(team.id, 0)
            if count < minCount:
                minCount = count
                minTeams = [team.id]
            elif count == minCount:
                minTeams.append(team.id)
        return random.choice(minTeams)


class GameStateBase(Static):

    startsNewMatch = False
    inMatch = True
    statsCount = False
    userTitle = ''
    userInfo = ()

    def applyTo(self, world):
        pass

    def canShoot(self):
        return True

    def becomeNeutralWhenDie(self):
        return False

    def getTeamToJoin(self, preferredTeamId, universe):
        return preferredTeamOtherwiseSmallest(preferredTeamId, universe)

    def aisEnabled(self):
        return False

    def canRename(self):
        return False

    def getGameStarter(self, universe, gameStartMethod):
        return None

    def isStartingCountdown(self):
        return False

    def displayWinner(self):
        return False

    def getGameTimeMode(self):
        return 'running'

    def isTrosball(self):
        return False

    def getInProgressState(self):
        return InProgress

    def getFinishedMessage(self, world):
        return None

    def onBecomingCurrentState(self, universe):
        pass

    def isTimedGameState(self):
        return False

    def getInfo(self, world, player):
        '''
        Returns information to display to the user about this game state.
        '''
        return self.userTitle, self.userInfo


@registerGameState
class Lobby(GameStateBase):
    id = 'Lobby'
    inMatch = False

    userTitle = 'Lobby'
    userInfo = (
        '* Free for all',
        '* Select preferred options from top menu',
        '* A match will start when enough players select "ready"',
    )

    def applyTo(self, world):
        super(Lobby, self).applyTo(world)
        world.abilities.zoneCaps = False
        world.abilities.stars = False
        world.onChangeVoiceChatRooms([], [])

    def becomeNeutralWhenDie(self):
        return True

    def getTeamToJoin(self, preferredTeamId, universe):
        return NeutralTeamId

    def aisEnabled(self):
        return True

    def canRename(self):
        return True

    def getGameStarter(self, universe, gameStartMethod):
        return VoteArbiter(universe, gameStartMethod)

    def getGameTimeMode(self):
        return 'unstarted'

    def getInfo(self, world, player):
        '''
        Returns information to display to the user about this game state.
        '''
        if world.canVote():
            return self.userTitle, self.userInfo
        return 'Game Over', ('Use the menu in the bottom left to leave game',)


@registerGameState
class Starting(GameStateBase):
    id = 'Starting'
    startsNewMatch = True

    userTitle = 'Get Ready...'
    userInfo = (
        '* Game will begin soon',
        '* Capture or neutralise all enemy zones',
        '* To capture a zone, touch the orb',
        "* If a team's territory is split, the smaller section is neutralised",
    )

    def applyTo(self, world):
        super(Starting, self).applyTo(world)
        world.abilities.upgrades = False
        world.abilities.respawn = False
        world.abilities.leaveFriendlyZones = False
        world.abilities.zoneCaps = False
        world.abilities.stars = False
        world.onChangeVoiceChatRooms(world.teams, world.players)

    def isStartingCountdown(self):
        return True

    def getGameTimeMode(self):
        return 'unstarted'


class StartingSolo(Starting):
    def aisEnabled(self):
        return True

    def getInProgressState(self):
        return Solo


@registerGameState
class StartingTrosball(Starting):
    id = 'StartingTrosball'

    userInfo = (
        '* Game will begin soon',
        '* Score points by getting the trosball through the net',
        '* To throw the trosball, press the "use upgrade" key',
        '* The trosball explodes if held for too long',
    )

    def applyTo(self, world):
        super(StartingTrosball, self).applyTo(world)
        world.uiOptions.showNets = True
        world.uiOptions.shiftingFrontLine = True
        world.uiOptions.frontLineGetter = (
            lambda: world.getTrosballPosition()[0])
        world.onChangeVoiceChatRooms(world.teams, world.players)

    def isTrosball(self):
        return True

    def getInProgressState(self):
        return Trosball

    def onBecomingCurrentState(self, universe):
        universe.resetTrosballToCentreOfMap()


@registerGameState
class Trosball(GameStateBase):
    id = 'Trosball'
    statsCount = True

    userTitle = 'Trosball'
    userInfo = (
        '* Score points by getting the trosball through the net',
        '* To throw the trosball, press the "use upgrade" key',
        '* The trosball explodes if held for too long',
    )

    def applyTo(self, world):
        super(Trosball, self).applyTo(world)
        world.abilities.zoneCaps = False
        world.abilities.stars = False

        world.uiOptions.showNets = True
        world.uiOptions.shiftingFrontLine = True
        world.uiOptions.frontLineGetter = (
            lambda: world.getTrosballPosition()[0])

    def isTrosball(self):
        return True

    def getFinishedMessage(self, world):
        if world.timer.hasTimeLimit() and world.timer.getTimeLeft() <= 0:
            teams = world.teamsWithHighestTrosballScore()
            teamId = '\x00' if len(teams) != 1 else teams[0].id
            return GameOverMsg(teamId, True)
        tagDistance = 35
        for team in world.teams:
            distance = math.distance(world.getTrosballPosition(),
                    world.getTrosballTargetZoneDefn(team).pos)
            if distance < tagDistance:
                playerId = (world.trosballPlayer.id if
                        world.trosballPlayer is not None else '\x00')
                return TrosballTagMsg(team.id, playerId)

    def isTimedGameState(self):
        return True

@registerGameState
class BetweenTrosballPoints(Trosball):
    id = 'BetweenTrosballPoints'
    statsCount = True

    def getFinishedMessage(self, world):
        return None


class StartingSoloTrosball(StartingTrosball):
    def aisEnabled(self):
        return True

    def getInProgressState(self):
        return SoloTrosball

class SoloTrosball(Trosball):
    def aisEnabled(self):
        return True

    def canRename(self):
        return True

    def isTimedGameState(self):
        return True

@registerGameState
class InProgress(GameStateBase):
    id = 'InProgress'
    statsCount = True

    userTitle = 'Trosnoth Match'
    userInfo = (
        '* Capture or neutralise all enemy zones',
        '* To capture a zone, touch the orb',
        "* If a team's territory is split, the smaller section is neutralised",
    )

    def getFinishedMessage(self, world):
        # Check first for timeout
        if world.timer.hasTimeLimit() and world.timer.getTimeLeft() <= 0:
            team = world.teamWithMoreZones()
            if team is not None:
                return GameOverMsg(team.id, True)
            else:
                return GameOverMsg(NeutralTeamId, True)
        # Next check for an actual winning team
        else:
            team = world.teamWithAllZones()
            if team == 'Draw':
                return GameOverMsg(NeutralTeamId, False)
            elif team is not None:
                return GameOverMsg(team.id, False)
            else:
                # No result yet
                pass
        return None

    def isTimedGameState(self):
        return True


@registerGameState
class Solo(InProgress):
    id = 'Solo'
    def aisEnabled(self):
        return True

    def canRename(self):
        return True

    def isTimedGameState(self):
        return True


@registerGameState
class Ended(GameStateBase):
    id = 'Ended'

    userTitle = 'Rabbit Hunt'

    def applyTo(self, world):
        super(Ended, self).applyTo(world)
        world.abilities.zoneCaps = False
        world.uiOptions.showWinner = True
        world.onChangeVoiceChatRooms([], [])

    def becomeNeutralWhenDie(self):
        return True

    def getTeamToJoin(self, preferredTeamId, universe):
        return NeutralTeamId

    def getGameTimeMode(self):
        return 'ended'

    def getInfo(self, world, player):
        '''
        Returns information to display to the user about this game state.
        '''
        userInfo = []
        winner = world.getWinningTeam()
        if winner is None:
            userInfo.append('* The game was a draw')
        else:
            userInfo.append('* %s win!' % (winner.teamName,))

        if player is None or player.team is None:
            if world._onceOnly:
                userInfo.extend([
                    '* Players who survived are "rabbits"',
                    '* Try to shoot all the rabbits',
                ])
            else:
                userInfo.extend([
                    '* Players who survived are "rabbits"',
                    '* Try to shoot all the rabbits before the time runs out',
                ])
        elif world._onceOnly:
            userInfo.extend([
                '* Try to survive for as long as possible',
                '* Try to kill any surviving enemy rabbits',
                '* If you die, you will respawn as a rogue',
            ])
        else:
            userInfo.extend([
                '* Try to survive until the time runs out',
                '* Try to kill any surviving enemy rabbits',
                '* If you die, you will respawn as a rogue',
            ])
        return self.userTitle, userInfo



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

class Team(object):
    '''Represents a team of the game'''
    def __init__(self, world, teamID):
        self.world = world
        self.numZonesOwned = 0
        self.orbScore = 0

        self.usingMinimapDisruption = False

        if (not isinstance(teamID, str)) or len(teamID) != 1:
            raise TypeError, 'teamID must be a single character'
        self.id = teamID

        if teamID == 'A':
            self.teamName = 'Blue players'
        elif teamID == 'B':
            self.teamName = 'Red players'
        else:
            self.teamName = '%s Team' % (teamID,)

    def __str__(self):
        return self.teamName

    def zoneLost(self):
        '''Called when a orb belonging to this team has been lost'''
        self.numZonesOwned -= 1
        self.opposingTeam.gainPoint()

    def zoneGained(self, score=0):
        '''Called when a orb has been attributed to this team'''
        self.numZonesOwned += 1
        self.gainPoint()

    def resetScore(self):
        self.orbScore = 0
        self.world.onTeamScoreChanged()

    def gainPoint(self):
        self.orbScore += 1
        self.world.onTeamScoreChanged()

    def isLoser(self):
        '''Returns whether the team has lost the game (True = Lost Game)'''
        return self.numZonesOwned == 0

    @staticmethod
    def setOpposition(teamA, teamB):
        teamA.opposingTeam = teamB
        teamB.opposingTeam = teamA



'''
Provides a layer between a universe and the GUI, turning
players, shots, grenades into sprites, and drawing mapblocks.
'''

import logging

from trosnoth.trosnothgui.ingame.sprites import (
    PlayerSprite, ShotSprite, GrenadeSprite, ExplosionSprite,
    CollectableStarSprite, ShoxwaveExplosionSprite, TrosballSprite,
    TrosballExplosionSprite)
from trosnoth.utils import globaldebug

log = logging.getLogger(__name__)


class UniverseGUI(object):

    # For debugging, displays where the player is server-side too
    SHOW_SERVER_SHADOWS = False

    def __init__(self, app, gameViewer, universe):
        self.app = app
        self.gameViewer = gameViewer
        self.universe = universe
        self.playerSprites = {}     # playerId -> PlayerSprite
        self.localPlayerId = None
        self.localPlayerSprite = None
        self.shotSprites = {}       # shotId -> ShotSprite
        self.localShotSprites = {}  # shot -> ShotSprite
        self.grenadeSprites = {}    # playerId -> GrenadeSprite
        self.localGrenadeSprite = None
        self.collectableStarSprites = {}    # starId -> CollectableStarSprite
        self.extraSprites = set()
        self.trosballSprite = None
        self.tweenFraction = 1

        app.displaySettings.onDetailLevelChanged.addListener(
            self.detailLevelChanged)

    def stop(self):
        self.app.displaySettings.onDetailLevelChanged.removeListener(
            self.detailLevelChanged)

    def detailLevelChanged(self):
        # Clear everything long-lived
        self.playerSprites = {}
        self.collectableStarSprites = {}
        self.trosballSprite = None
        if self.localPlayerId is not None:
            p = self.localPlayerSprite
            self.localPlayerSprite = PlayerSprite(self.app, self, p.player)

    def setTweenFraction(self, f):
        self.tweenFraction = f

    @property
    def zones(self):
        return self.universe.zones

    @property
    def teams(self):
        return self.universe.teams

    @property
    def map(self):
        return self.universe.map

    @property
    def zoneBlocks(self):
        return self.universe.zoneBlocks

    def getTrosballSprite(self):
        try:
            self.universe.getTrosballPosition()
        except ValueError:
            return None
        if self.trosballSprite is None:
            self.trosballSprite = TrosballSprite(self.app, self, self.universe)
        return self.trosballSprite

    def getPlayerSprite(self, playerId, ignoreLocal=False):
        player = self.universe.getPlayer(playerId)
        if player is None:
            return None
        if playerId == self.localPlayerId and not ignoreLocal:
            p = self.localPlayerSprite
            if p.spriteTeam != p.player.team:
                self.localPlayerSprite = PlayerSprite(
                    self.app, self, p.player)
                if p is self.gameViewer.viewManager.target:
                    self.gameViewer.viewManager.setTarget(
                        self.localPlayerSprite)
                p = self.localPlayerSprite
            return p

        try:
            p = self.playerSprites[playerId]
        except KeyError:
            self.playerSprites[player.id] = p = PlayerSprite(
                self.app, self, player)
            return p

        if p.spriteTeam != player.team:
            # Player has changed teams.
            self.playerSprites[player.id] = p = PlayerSprite(
                self.app, self, player)
        return p

    def overridePlayer(self, player):
        self.localPlayerId = player.id
        self.localPlayerSprite = PlayerSprite(self.app, self, player)

    def removeOverride(self):
        self.localPlayerId = None
        self.localPlayerSprite = None

    def iterPlayers(self):
        untouched = set(self.playerSprites.iterkeys())
        for player in self.universe.players:
            untouched.discard(player.id)
            yield self.getPlayerSprite(player.id)
            if self.SHOW_SERVER_SHADOWS and player.id == self.localPlayerId:
                yield self.getPlayerSprite(player.id, ignoreLocal=True)

        # Clean up old players.
        for playerId in untouched:
            del self.playerSprites[playerId]

    def iterGrenades(self):
        untouched = set(self.grenadeSprites.iterkeys())
        for grenade in self.universe.grenades:
            if (
                    grenade.player.id == self.localPlayerId
                    and not self.SHOW_SERVER_SHADOWS):
                continue
            try:
                yield self.grenadeSprites[grenade.player.id]
            except KeyError:
                self.grenadeSprites[grenade.player.id] = g = (
                    GrenadeSprite(self.app, self, grenade))
                yield g
            else:
                untouched.discard(grenade.player.id)

        grenade = self.gameViewer.interface.localState.localGrenade
        if grenade:
            if self.localGrenadeSprite is None:
                self.localGrenadeSprite = GrenadeSprite(
                    self.app, self, grenade)
            yield self.localGrenadeSprite
        else:
            self.localGrenadeSprite = None

        # Clean up old grenades.
        for grenadeId in untouched:
            del self.grenadeSprites[grenadeId]

    def iterShots(self):
        untouched = set(self.shotSprites.iterkeys())
        untouchedLocals = set(self.localShotSprites.iterkeys())
        for shot in self.universe.shots:
            if (
                    shot.originatingPlayer.id == self.localPlayerId
                    and not self.SHOW_SERVER_SHADOWS):
                continue
            if shot.justFired:
                continue
            if shot.expired and not globaldebug.enabled:
                continue
            try:
                yield self.shotSprites[shot.id]
            except KeyError:
                self.shotSprites[shot.id] = s = (
                    ShotSprite(self.app, self, shot))
                yield s
            else:
                untouched.discard(shot.id)

        for shot in self.gameViewer.interface.localState.shots:
            if shot.justFired:
                continue
            try:
                yield self.localShotSprites[shot]
            except KeyError:
                self.localShotSprites[shot] = s = ShotSprite(
                    self.app, self, shot)
                yield s
            else:
                untouchedLocals.discard(shot.id)

        # Clean up old shots.
        for shotId in untouched:
            del self.shotSprites[shotId]
        for shot in untouchedLocals:
            del self.localShotSprites[shot]

    def iterCollectableStars(self):
        untouched = set(self.collectableStarSprites.iterkeys())
        for star in self.universe.collectableStars.itervalues():
            if star.hitLocalPlayer:
                continue
            try:
                yield self.collectableStarSprites[star.id]
            except KeyError:
                self.collectableStarSprites[star.id] = s = (
                    CollectableStarSprite(self.app, self, star))
                yield s
            else:
                untouched.discard(star.id)

        # Clean up old shots.
        for starId in untouched:
            del self.collectableStarSprites[starId]

    def iterExtras(self):
        for sprite in list(self.extraSprites):
            if sprite.isDead():
                self.extraSprites.remove(sprite)
            yield sprite
        trosball = self.getTrosballSprite()
        if trosball:
            yield trosball

    def getPlayerCount(self):
        return len(self.universe.players)

    def hasPlayer(self, player):
        return (
            player.id in self.playerSprites or player.id == self.localPlayerId)

    def getPlayersInZone(self, zone):
        result = []
        for p in zone.players:
            ps = self.getPlayerSprite(p.id)
            if ps is not None:
                result.append(ps)
        return result

    def addExplosion(self, pos):
        self.extraSprites.add(ExplosionSprite(self.app, pos))

    def addTrosballExplosion(self, pos):
        self.extraSprites.add(TrosballExplosionSprite(self.app, pos))

    def addShoxwaveExplosion(self, pos):
        self.extraSprites.add(ShoxwaveExplosionSprite(self.app, pos))

    def getTrosballPosition(self):
        return self.universe.getTrosballPosition()



'''interface.py - defines the Interface class which deals with drawing the
game to the screen, including menus and other doodads.'''

import logging

import pygame
from twisted.internet import defer, reactor
from twisted.internet.protocol import ClientCreator

from trosnoth.network.client import TrosnothClientProtocol, ConnectionFailed
from trosnoth.game import RemoteGame
from trosnoth.gamerecording import replays

from trosnoth.gui.framework import framework
from trosnoth.model.uithrottler import UIMsgThrottler, LocalGameTweener

from trosnoth.trosnothgui.pregame.startup import StartupInterface
from trosnoth.trosnothgui.pregame.connectingScreen import ConnectingScreen
from trosnoth.trosnothgui.pregame.backdrop import TrosnothBackdrop
from trosnoth.trosnothgui.pregame.playscreen import PlayAuthScreen

from trosnoth.trosnothgui.pregame.connectionFailedDialog import (
    ConnectionFailedDialog)
from trosnoth.trosnothgui.ingame.gameInterface import GameInterface
from trosnoth.trosnothgui.pregame.savedGameMenu import SavedGameMenu
from trosnoth.trosnothgui.settings.settings import SettingsMenu
import trosnoth.version

log = logging.getLogger('interface')


class Interface(framework.CompoundElement):
    def __init__(self, app):
        super(Interface, self).__init__(app)

        # Create an interfaces for pre- and post-connection.
        self.backdrop = TrosnothBackdrop(app)

        self.gi = None
        self.game = None
        self.gameIsLocal = False
        self.currentDeferred = None
        self.trosnothClient = None
        self.startupInterface = self._makeStartupInterface()

        self.elements = [self.backdrop, self.startupInterface]

    def _makeStartupInterface(self):
        result = StartupInterface(self.app, self)
        result.start()
        return result

    def processEvent(self, event):
        # Capture the quit event.
        if event.type == pygame.QUIT:
            self.app.stop()
            return
        return super(Interface, self).processEvent(event)

    @defer.inlineCallbacks
    def connectToServer(self, host, port, timeout=7):
        self.elements = [
            self.backdrop,
            ConnectingScreen(
                self.app, '%s:%s' % (host, port),
                onCancel=self.cancelConnecting)
        ]

        try:
            cc = ClientCreator(reactor, TrosnothClientProtocol)
            self.currentDeferred = cc.connectTCP(host, port, timeout=timeout)
            trosnothClient = yield self.currentDeferred
            self.trosnothClient = trosnothClient
            self.currentDeferred = trosnothClient.getSettings()
            settings = yield self.currentDeferred
            self.currentDeferred = None
            self.connectionEstablished(settings)

        except Exception as e:
            self.currentDeferred = None
            self.trosnothClient = None
            self.elements = [self.backdrop, self.startupInterface]
            if not isinstance(e, defer.CancelledError):
                if isinstance(e, ConnectionFailed):
                    text = str(e.reason)
                else:
                    text = 'Internal Error'
                    log.exception('Unexpected failure in deferred')
                d = ConnectionFailedDialog(self.app, text)
                d.show()

    def cancelConnecting(self, source):
        if self.currentDeferred:
            self.currentDeferred.cancel()

    def connectedToGame(self, trosnothClient, settings, authTag):
        self.trosnothClient = trosnothClient
        self.connectionEstablished(settings, authTag)

    def connectToLocalServer(self):
        '''
        Requries app.server is not None. Connects to the locally-hosted server.
        '''
        self.connectToGameObject(self.app.server.game)

    def connectToGameObject(self, game):
        self.game = game
        self.app.tweener = LocalGameTweener(game)
        self.setupGame(game, 0, local=True)

        self.elements = [self.gi]

    def connectionEstablished(self, settings, authTag=0):
        'Called when this client establishes a connection to a server.'
        self.trosnothClient.onConnectionLost.addListener(self.connectionLost)

        game = RemoteGame(self.app.layoutDatabase)
        self.app.tweener = UIMsgThrottler()
        self.trosnothClient.connectNode(self.app.tweener)
        self.app.tweener.connectNode(game)

        self.setupGame(game, authTag, local=False)
        game.connected(settings)

        self.game = game
        self.elements = [self.gi]

    def setupGame(self, game, authTag, local):
        # Create a game interface.
        self.gi = GameInterface(
            self.app, game,
            onDisconnectRequest=self._disconnectGame,
            onConnectionLost=self.connectionLost, authTag=authTag)
        self.gameIsLocal = local

    def setupReplay(self, game):
        self.gi = GameInterface(
            self.app, game, onDisconnectRequest=self.replayOver, replay=True,
            onConnectionLost=self.replayOver)

    def replayOver(self):
        self.elements = [self.backdrop, self.startupInterface]

    def _disconnectGame(self):
        '''
        Requested to disconnect from the game.
        '''
        if self.trosnothClient is not None:
            self.trosnothClient.disconnect()
            self.trosnothClient.disconnectNode()
            self.trosnothClient = None
        else:
            self.connectionLost()

    def connectionLost(self, reason=None):
        self.elements = [self.backdrop, self.startupInterface]
        if self.game is not None:
            if self.gi:
                self.game.detachAgent(self.gi)
            if not self.gameIsLocal:
                # Do not stop a local server, in case other people are still
                # playing.
                self.game.stop()
            self.game = None
        if self.gi is not None:
            self.gi.stop()
            self.gi = None

        if reason is not None:
            d = ConnectionFailedDialog(self.app, reason)
            d.show()

    def connectToReplay(self, filename):
        self.trosnothClient = player = replays.ReplayPlayer(filename)

        game = RemoteGame(self.app.layoutDatabase)
        self.app.tweener = UIMsgThrottler()

        player.connectNode(self.app.tweener)
        self.app.tweener.connectNode(game)

        self.setupReplay(game)
        game.connected(player.popSettings())
        player.start()

        self.game = game
        self.elements = [self.gi]


class SingleAuthInterface(Interface):
    connectScreenFactory = PlayAuthScreen

    def __init__(self, app):
        Interface.__init__(self, app)

        if self.app.serverDetails:
            self.startupInterface.begin(
                [self.app.serverDetails, 'create'], canHost=False)
        else:
            settings = app.connectionSettings
            servers = list(settings.servers)
            if not trosnoth.version.release:
                # Check that developers aren't accidentally connecting to
                # play.trosnoth.org.
                servers = [
                    s[:2] for s in servers if s[0] != 'play.trosnoth.org']
            if settings.otherGames:
                servers.append('others')
            if settings.lanGames == 'afterInet':
                servers.append('lan')
            elif settings.lanGames == 'beforeInet':
                servers.insert(0, 'lan')
            if settings.createGames:
                servers.append('create')
            servers.insert(0, 'self')
            self.startupInterface.begin(servers, canHost=True)

    def _makeStartupInterface(self):
        return self.connectScreenFactory(self.app, onFail=self.app.stop)

    def connectionLost(self, reason=None):
        Interface.connectionLost(self, reason)
        self.app.stop()


class KeySettingsInterface(Interface):
    def _makeStartupInterface(self):
        return SettingsMenu(
            self.app, onClose=self.app.stop, showSound=False,
            showDisplay=False)


class ArchivesInterface(Interface):
    def _makeStartupInterface(self):
        return SavedGameMenu(
            self.app, onCancel=self.app.stop, onReplay=self.connectToReplay)



DEFAULT_TEAM_NAME_1 = 'Defaulticons'
DEFAULT_TEAM_NAME_2 = 'Standardators'
STALEMATE_CHECK_PERIOD = 1
COLLECTABLE_STAR_LIFETIME = 12
NO_COLLECTABLE_STARS = False        # For debugging.
TICK_PERIOD = 0.05
MAX_PLAYER_NAME_LENGTH = 30

# The time after a zone tag after which there is 50% of a new
# collectable star appearing every STALEMATE_CHECK_PERIOD.
ZONE_TAG_STAR_HALF_LIFE = 60

GAME_STATE_CHECK_PERIOD = 5
RABBIT_CHASING_PERIOD = 60


MAP_TO_SCREEN_SCALE = 1.25

BODY_BLOCK_MAP_SIZE = (1024, 384)
BODY_BLOCK_SCREEN_SIZE = (
    int(BODY_BLOCK_MAP_SIZE[0] * MAP_TO_SCREEN_SCALE + 0.5),
    int(BODY_BLOCK_MAP_SIZE[1] * MAP_TO_SCREEN_SCALE + 0.5),
)
INTERFACE_BLOCK_MAP_SIZE = (512, 384)
INTERFACE_BLOCK_SCREEN_SIZE = (
    int(INTERFACE_BLOCK_MAP_SIZE[0] * MAP_TO_SCREEN_SCALE + 0.5),
    int(INTERFACE_BLOCK_MAP_SIZE[1] * MAP_TO_SCREEN_SCALE + 0.5),
)


# When a client is found to be lagging, this is the number of extra frames of
# delay that's added (on top of the measured lag), to prevent lots of resyncs
# in quick succession.
LAG_BUFFER = 1
LAG_THRESHOLD = 5
INITIAL_ASSUMED_LATENCY = int(0.3 / TICK_PERIOD)


PATH_FINDING_DISCRETISATION_UNIT = 4.5



# deathTypes
SHOT_DEATH = 'S'
TURRET_DEATH = 'T'
GRENADE_DEATH = 'G'
SHOXWAVE_DEATH = 'W'
OFF_MAP_DEATH = 'O'
TROSBALL_DEATH = 'B'
BOMBER_DEATH = 'X'


# ChatMsg kinds
OPEN_CHAT = 'o'
PRIVATE_CHAT = 'p'
TEAM_CHAT = 't'


# CannotBuyUpgradeMsg reasons
NOT_ENOUGH_STARS_REASON = 'N'
HAVE_UPGRADE_REASON = 'A'
PLAYER_DEAD_REASON = 'D'
GAME_NOT_STARTED_REASON = 'P'
ALREADY_TURRET_REASON = 'T'
TOO_CLOSE_TO_EDGE_REASON = 'E'
TOO_CLOSE_TO_ORB_REASON = 'O'
NOT_IN_DARK_ZONE_REASON = 'F'
ALREADY_DISRUPTED_REASON = 'M'
INVALID_UPGRADE_REASON = 'I'
DISABLED_UPGRADE_REASON = 'X'


# CannotJoinMsg reasons
GAME_FULL_REASON = 'F'
UNAUTHORISED_REASON = 'A'
NICK_USED_REASON = 'N'
BAD_NICK_REASON = 'B'
USER_IN_GAME_REASON = 'U'
ALREADY_JOINED_REASON = 'J'


# CannotRespawnMsg reasons
RESYNC_IN_PROGRESS_REASON = 'r'
# GAME_NOT_STARTED_REASON = 'P' (defined above)
ALREADY_ALIVE_REASON = 'A'
BE_PATIENT_REASON = 'T'
ENEMY_ZONE_REASON = 'E'
FROZEN_ZONE_REASON = 'F'



from trosnoth.model.zone import ZoneDef, ZoneState
from trosnoth.utils.unrepr import unrepr


class MapLayout(object):
    '''Stores static info about the layout of the map.

    Attributes:
        centreX, centreY: the x and y coordinates of the map centre.
        zones: collection of zoneDefs
        blocks: collection of blockDefs
    '''

    # The dimensions of zones. See diagram below.
    halfZoneHeight = 384        # a / 2
    zoneBodyWidth = 1024        # b
    zoneInterfaceWidth = 512    # c

    # This diagram explains the dimensions defined above.
    #     \___________/ _ _ _ _
    #     /           \       ^
    #    /             \      |
    # __/               \___  a
    #   \               /|    |
    #    \             / |    |
    #     \___________/ _|_ _ v
    #     /|         |\  |
    #      |<-- b -->|<c>|

    def __init__(self, halfMapWidth, mapHeight):
        self.halfMapWidth = halfMapWidth
        self.mapHeight = mapHeight
        self.pathFinder = None

        self.trosballTargetZones = [None, None]

        from trosnoth.model.mapblocks import MapBlockDef
        totalWidth = halfMapWidth * 2 + 1
        middleHeight = ((halfMapWidth - 1) % 2) + 1 + mapHeight

        self.worldSize = (
            totalWidth * self.zoneBodyWidth + (
                totalWidth + 1) * self.zoneInterfaceWidth,
            2 * middleHeight * self.halfZoneHeight
        )

        if middleHeight == mapHeight + 2:
            highExists = True
        else:
            highExists = False

        # Calculate position of centre.
        self.centreX = (
            (halfMapWidth + 0.5) * MapLayout.zoneBodyWidth
            + (halfMapWidth + 1) * MapLayout.zoneInterfaceWidth)
        self.centreY = middleHeight * MapLayout.halfZoneHeight

        # Collection of all zone definitions:
        self.zones = set()

        # Zones shall also be stored as a multidimensional array
        # in the following format of column/vertical position:
        #    0   1   2   3   4   5   6
        # 0 -/a  a  a\-  -  -/d  d  d\-
        # 1 -\a  a  a/c  c  c\d  d  d/-
        # 2 -/b  b  b\c  c  c/e  e  e\-
        # 3 -\b  b  b/-  -  -\e  e  e/-
        # so that the players' current zone can more easily be calculated (in
        # the getMapBlockIndices procedure below)

        # This will be stored in the format of a list of lists of MapBlocks.
        # The most encompassing list will store the rows.
        # Each "row" will be a list of blocks.
        # Each "block" will be an instance of a subclass of MapBlock.

        self.blocks = []
        blockTypes = ('btm', 'fwd', 'top', 'bck')
        y = 0

        # Decide which type of block to start with.
        if halfMapWidth % 2 == 0:
            nextBlock = 1
        else:
            nextBlock = 3

        # Initialise self.blocks
        for i in range(middleHeight * 2):
            row = []
            x = 0
            bodyBlock = None
            for j in range(totalWidth):
                # Add an interface.
                blockType = blockTypes[nextBlock]
                ifaceBlock = MapBlockDef(blockType, x, y)
                row.append(ifaceBlock)

                x = x + MapLayout.zoneInterfaceWidth
                nextBlock = (nextBlock + 1) % 4

                # Add a body block.
                blockType = blockTypes[nextBlock]
                bodyBlock = MapBlockDef(blockType, x, y)
                row.append(bodyBlock)

                x = x + MapLayout.zoneBodyWidth
                nextBlock = nextBlock + 1

            # Add the last interface.
            blockType = blockTypes[nextBlock]
            ifaceBlock = MapBlockDef(blockType, x, y)
            row.append(ifaceBlock)

            self.blocks.append(row)
            y = y + MapLayout.halfZoneHeight

        # height: low means the column is of height mapHeight
        # mid means it is mapHeight + 1,
        # high means mapHeight + 2 (middleHeight)

        # height starts (and ends) at low in every case.
        height = "low"

        # numSoFar keeps track of the number of zones created so far.
        numSoFar = 0

        ''' To calculate adjacent zones, the zones were conceptualised in the
         following pattern, counting downwards by column, then left to right:
         (assuming mapHeight = 3, halfMapWidth = 2)
              7
            3    12
         0    8     16
            4    13
         1    9     17
            5    14
         2    10    18
            6    15
              11

        '''

        '''In order to create each of the zones, we will iterate firstly
        through each of the columns, then through each of the zones in each
        column. This allows us to calculate adjacent zones based on column
        height.'''

        # Keep track of where the zones should be put.
        xCoord = MapLayout.zoneInterfaceWidth + int(
            MapLayout.zoneBodyWidth / 2)

        # Iterate through each column:
        for i in range(0, totalWidth):
            startCol = i * 2

            # Determine number of zones in the current column
            if height == "low":
                colHeight = mapHeight
            elif height == "mid":
                colHeight = mapHeight + 1
            elif height == "high":
                colHeight = mapHeight + 2

            # Calculate y-coordinate of the first zone in the column.
            if height == "low" and highExists:
                yCoord = MapLayout.halfZoneHeight
            elif ((height == "mid" and highExists) or (
                    height == "low" and not highExists)):
                yCoord = 0
            elif height == "high" or (height == "mid" and not highExists):
                yCoord = -MapLayout.halfZoneHeight

            prevZone = None

            # Iterate through each zone in the column
            for j in range(0, colHeight):

                # Determine Team, and create a zone.
                if i < halfMapWidth:
                    team = 0
                elif i > (halfMapWidth):
                    team = 1
                else:
                    team = None

                # Determine co-ordinates of the zone's orb (middle of zone)
                yCoord = yCoord + 2 * MapLayout.halfZoneHeight

                newZone = ZoneDef(numSoFar, team, xCoord, yCoord)

                if j == (colHeight - 1) // 2:
                    # The middle zone, height-wise
                    if i == 0:
                        # On the left is team 2's target
                        self.trosballTargetZones[1] = newZone
                    elif i == totalWidth - 1:
                        # On the right is team 1's target
                        self.trosballTargetZones[0] = newZone

                # Link this zone to the zone above.
                if prevZone is not None:
                    connectingBlocks = (
                        self.blocks[y][startCol + 1],
                        self.blocks[y + 1][startCol + 1])
                    prevZone.adjacentZones[newZone] = connectingBlocks
                    newZone.adjacentZones[prevZone] = connectingBlocks

                # Add zone to list of zones
                self.zones.add(newZone)

                # Add zone to blocks
                if highExists and height == "low":
                    startRow = 2 + j * 2
                elif (highExists and height == "mid") or (
                        not highExists and height == "low"):
                    startRow = 1 + j * 2
                elif height == "high" or (not highExists and height == "mid"):
                    startRow = 0 + j * 2
                else:
                    raise Exception("There's an error in blocks allocation")

                # Put new zone into blocks, and connect it to zones left.
                for y in range(startRow, startRow + 2):
                    cornerBlock = self.blocks[y][startCol + 2]
                    cornerBlock.zone1 = newZone
                    bodyBlock = self.blocks[y][startCol + 1]
                    bodyBlock.zone = newZone
                    block = self.blocks[y][startCol]
                    block.zone2 = newZone
                    newZone.bodyBlocks.append(bodyBlock)
                    newZone.interfaceBlocks.extend([cornerBlock, block])

                    if block.zone1:
                        # Connect this block to the block to the left.
                        block.zone1.adjacentZones[newZone] = (block,)
                        newZone.adjacentZones[block.zone1] = (block,)

                # If block is at the top of a column, it must have a roof.
                if prevZone is None:
                    self.blocks[startRow][startCol + 1].blocked = True

                numSoFar = numSoFar + 1
                prevZone = newZone

            # Make sure that the bottom zone in each column has a floor.
            self.blocks[startRow + 1][startCol + 1].blocked = True

            # Advance x-coordinate.
            xCoord = (
                xCoord + MapLayout.zoneBodyWidth +
                MapLayout.zoneInterfaceWidth)

            # Determine how high the next column will be:
            if height == "low":
                if i == totalWidth - 1:
                    height = None
                else:
                    height = "mid"
            elif height == "mid":
                if i == halfMapWidth - 1:
                    height = "high"
                else:
                    height = "low"
            elif height == "high":
                height = "mid"

        for row in self.blocks:
            for i in xrange(0, len(row), 2):
                block = row[i]
                if block.zone1 is None or block.zone2 is None:
                    # Make sure the edges are blocked.
                    block.blocked = True
        # Zone initialisation and calculations end here.

    def getTrosballTargetZoneDefn(self, team):
        if team.id == 'A':
            return self.trosballTargetZones[0]
        elif team.id == 'B':
            return self.trosballTargetZones[1]

    def toString(self):
        '''
        Serialises this MapLayout into a string that can be sent over the
        network, then reconstructed with fromString().
        '''
        result = []
        for blockRow in self.blocks:
            row = []
            result.append(row)
            for block in blockRow:
                if block.layout is None:
                    row.append(None)
                else:
                    row.append(block.layout.key)
        return repr([(self.halfMapWidth, self.mapHeight), result])

    @staticmethod
    def unknownBlockKeys(layoutDatabase, text):
        '''
        Goes through a MapLayout's toString() representation and checks for and
        returns a collection of map block keys for all map blocks not known by
        the given layoutDatabase.
        '''
        result = set()
        data = unrepr(text)
        assert isinstance(data, list)
        discard, data = data
        for row in data:
            for key in row:
                if key is not None:
                    fKey = layoutDatabase.getFundamentalKey(key)
                    if fKey is not None:
                        result.add(fKey)

        return result

    @staticmethod
    def fromString(layoutDatabase, text):
        '''
        Reconstructs a MapLayout from its toString() result. Assumes that all
        map block definitions are available in the LayoutDatabase given.
        '''
        data = unrepr(text)
        assert isinstance(data, list)
        (halfMapWidth, mapHeight), data = data
        result = MapLayout(halfMapWidth, mapHeight)
        for i, row in enumerate(data):
            for j, key in enumerate(row):
                if key is not None:
                    layout = layoutDatabase.getLayoutByKey(key)
                    layout.applyTo(result.blocks[i][j])
        return result

    @staticmethod
    def getMapBlockIndices(xCoord, yCoord):
        '''Returns the index in Universe.zoneBlocks of the map block which the
        given x and y-coordinates belong to.
        (0, 0) is the top-left corner.

        To find a zone, use MapBlock.getZoneAtPoint()'''

        blockY = int(yCoord // MapLayout.halfZoneHeight)

        blockX, remainder = divmod(
            xCoord, MapLayout.zoneBodyWidth + MapLayout.zoneInterfaceWidth)
        if remainder >= MapLayout.zoneInterfaceWidth:
            blockX = int(2 * blockX + 1)
        else:
            blockX = int(2 * blockX)

        return blockY, blockX


class MapState(object):
    '''Stores dynamic info about the layout of the map including who owns what
    zone.

    Attributes:
        layout - the static MapLayout
        zones - a set of zones
        blocks - a set of blocks
        zoneWithId - mapping from zone id to zone
    '''

    def __init__(self, universe, layout):
        self.layout = layout

        self.zones = set()
        self.zoneWithDef = {}
        for zone in self.layout.zones:
            newZone = ZoneState(universe, zone)

            self.zones.add(newZone)
            self.zoneWithDef[zone] = newZone

        self.zoneWithId = {}
        for zone in self.zones:
            self.zoneWithId[zone.id] = zone

        self.zoneBlocks = []
        for row in self.layout.blocks:
            newRow = []
            for blockDef in row:
                newRow.append(blockDef.spawnState(universe, self.zoneWithDef))
            self.zoneBlocks.append(newRow)

    def getMapBlockAtPoint(self, pos):
        i, j = MapLayout.getMapBlockIndices(*pos)
        return self.zoneBlocks[i][j]

    def getZoneAtPoint(self, pos):
        try:
            return self.getMapBlockAtPoint(pos).getZoneAtPoint(*pos)
        except IndexError:
            return None



'''
Django settings for Trosnoth server.

For more information on this file, see
https://docs.djangoproject.com/en/1.7/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.7/ref/settings/
'''

import os

from trosnoth import data


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.7/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 't$gv+_7-d5de7j2s$(w%_8x8%&iyv$go9^5cl5_&u_+-hs$gn-'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'trosnoth.djangoapp.apps.TrosnothConfig',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'trosnoth.server.urls'

WSGI_APPLICATION = 'trosnoth.server.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.7/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': data.getPath(data.user, 'authserver', 'db.sqlite3'),
    }
}

# Internationalization
# https://docs.djangoproject.com/en/1.7/topics/i18n/

LANGUAGE_CODE = 'en-au'

USE_I18N = True

USE_L10N = True

USE_TZ = False

TIME_ZONE = None


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.7/howto/static-files/

STATIC_URL = '/static/'
STATICFILES_DIRS = [
    data.getPath(data.user, 'authserver', 'public'),
]



from optparse import OptionParser
import os
import sys

from trosnoth.run.main import Main
from trosnoth.utils.utils import initLogging


def makeOptionParser():
    parser = OptionParser()
    parser.add_option(
        '-u', '--auth-server', action='store', dest='server',
        help='connect to the given authentication server')
    parser.add_option(
        '-s', '--solo', action='store_const', dest='mode',
        const='solo', default='normal',
        help='run Trosnoth in solo mode.')
    parser.add_option(
        '-i', '--isolate', action='store_const', dest='mode', const='isolate',
        help='run Trosnoth in isolated test mode. Creates client and server '
        'Universe objects, but does not use the network to connect them.')
    parser.add_option(
        '-D', '--delay', action='store', dest='delay', type=float,
        help='implies --isolate. Adds a delay between client and server.')
    parser.add_option(
        '--trosball', action='store_const', dest='mode', const='trosball',
        help='run a solo Trosball game.')
    parser.add_option(
        '-t', '--test', action='store_true', dest='testMode',
        help='run Trosnoth in test mode. This means that players will get 20 '
        'stars each and upgrades will only cost 1 star.')
    parser.add_option(
        '-b', '--testblock', action='store', dest='mapBlock',
        help='tests the map block with the given filename.')
    parser.add_option(
        '-a', '--aicount', action='store', dest='aiCount',
        help='the number of AIs to include.')
    parser.add_option(
        '-w', '--halfwidth', action='store', dest='halfWidth',
        help='the half map width')
    parser.add_option(
        '-H', '--height', action='store', dest='height', help='the map height')
    parser.add_option(
        '-B', '--block-ratio', action='store', dest='blockRatio', default=0.5,
        type=float, help='the fraction of the map to block off')
    parser.add_option(
        '-m', '--duration', action='store', dest='duration', default=0,
        type=int, help='game duration in minutes')
    parser.add_option(
        '-S', '--stack-teams', action='store_true', dest='stackTeams',
        help='stack all the AI players on one team.')
    parser.add_option(
        '-A', '--aiclass', action='store', dest='aiClass',
        help='the name of the module to import from trosnoth.ais.')
    parser.add_option(
        '--list-ais', action='store_const', dest='mode',
        const='listais', help='list available AI classes and exit')
    parser.add_option(
        '-d', '--debug', action='store_true', dest='debug',
        help='show debug-level messages on console')
    parser.add_option(
        '-p', '--profile', action='store_true', dest='profile',
        help='dump kcachegrind profiling data to trosnoth.log')
    parser.add_option(
        '-r', '--log-reactor-calls', action='store_true',
        dest='log_reactor_calls', help='logs slow reactor calls')
    parser.add_option(
        '-P', '--profile-reactor-calls', action='store_true',
        dest='profile_reactor_calls', help='profiles slow reactor calls')
    parser.add_option(
        '-l', '--log-file', action='store', dest='logFile',
        help='file to write logs to')
    return parser


def listAIs():
    import trosnoth.ai
    print >> sys.stderr, 'Available AI classes:'
    for name in trosnoth.ai.listAIs():
        print >> sys.stderr, '  %s' % (name,)
    return


TESTMODE_AI_COUNT = 7
TESTMODE_HALF_WIDTH = 1
TESTMODE_HEIGHT = 1
BLOCKTEST_HALF_WIDTH = 2
BLOCKTEST_HEIGHT = 1


def processOptions(options, parser):
    if options.server is not None:
        if options.mode != 'normal':
            parser.error('multiple modes specified')
        options.mode = 'auth'

    if options.mode == 'normal' and options.delay:
        options.mode = 'isolate'
    if options.mode == 'normal' and (
            options.testMode or options.mapBlock or
            options.aiCount or options.halfWidth or options.height or
            options.stackTeams or options.aiClass or options.duration):
        options.mode = 'solo'

    if options.aiClass is None:
        options.aiClass = 'ranger'

    if options.mode == 'auth':
        if ':' in options.server:
            host, port = options.server.split(':', 1)
            try:
                port = int(port)
            except ValueError:
                parser.error('%r is not a valid port' % (port,))
        else:
            host = options.server
            port = 6787
        options.host, options.port = host, port
    elif options.mode in ('solo', 'isolate', 'trosball'):
        isolate = options.mode == 'isolate'
        trosball = options.mode == 'trosball'
        options.mode = 'solo'
        if options.mapBlock is None:
            mapBlocks = []
            aiCount = options.aiCount or TESTMODE_AI_COUNT
            halfWidth = options.halfWidth or TESTMODE_HALF_WIDTH
            height = options.height or TESTMODE_HEIGHT
            startInLobby = False
        else:
            mapBlocks = [options.mapBlock]
            aiCount = options.aiCount or 0
            halfWidth = options.halfWidth or BLOCKTEST_HALF_WIDTH
            height = options.height or BLOCKTEST_HEIGHT
            if (isolate or trosball or aiCount):
                startInLobby = False
            else:
                startInLobby = True
        options.soloArgs = dict(
            isolate=isolate, testMode=options.testMode,
            mapBlocks=mapBlocks, size=(int(halfWidth), int(height)),
            aiCount=int(aiCount), lobby=startInLobby,
            stackTeams=options.stackTeams, aiClass=options.aiClass,
            trosball=trosball, delay=options.delay,
            blockRatio=options.blockRatio,
            duration=options.duration,
        )

    elif options.mode not in ('normal', 'listais'):
        assert False, 'Unknown mode {!r}'.format(options.mode)


def main():
    parser = makeOptionParser()

    options, args = makeOptionParser().parse_args()
    options.replayFile = None
    if len(args) == 1:
        (replayFile,) = args
        if not os.path.exists(replayFile):
            parser.error('Replay file not found: {}'.format(replayFile))
        options.normal = 'replay'
        options.replayFile = replayFile
    elif len(args) > 1:
        parser.error('too many arguments')

    initLogging(options.debug, options.logFile)

    processOptions(options, parser)

    if options.mode == 'listais':
        listAIs()
        return

    if options.mode == 'normal':
        mainObject = Main(showReplay=options.replayFile)
    elif options.mode == 'auth':
        mainObject = Main(serverDetails=(options.host, options.port))
    elif options.mode == 'solo':
        from trosnoth.run import solotest
        mainObject = solotest.Main(**options.soloArgs)
    else:
        assert False, 'Unknown mode'

    if options.log_reactor_calls or options.profile_reactor_calls:
        from trosnoth.utils import twistdebug
        twistdebug.start(profiling=options.profile_reactor_calls)

    if options.profile:
        mainObject.run_with_profiling(twisted=True)
    else:
        mainObject.run_twisted()


if __name__ == '__main__':
    main()



import logging

from trosnoth.gui.framework import framework
from trosnoth.gui.framework.checkbox import CheckBox
from trosnoth.gui.framework.tab import Tab
from trosnoth.gui.framework.slider import Slider
from trosnoth.trosnothgui.common import button
from trosnoth.gui.common import ScaledLocation, ScaledArea
from trosnoth.gui.framework.elements import TextElement
from trosnoth.utils.event import Event

log = logging.getLogger()


class SoundSettingsTab(Tab, framework.CompoundElement):
    def __init__(self, app, onClose=None):
        super(SoundSettingsTab, self).__init__(app, "Sounds")

        self.onClose = Event()
        if onClose is not None:
            self.onClose.addListener(onClose)

        font = self.app.screenManager.fonts.bigMenuFont
        colours = app.theme.colours

        text = [
            TextElement(self.app, 'Music Volume', font,
                ScaledLocation(400, 285, 'topright'), colours.headingColour),
            TextElement(self.app, 'Enable Music', font,
                ScaledLocation(400, 355, 'topright'), colours.headingColour),
            TextElement(self.app, 'Sound Volume', font,
                ScaledLocation(400, 425, 'topright'), colours.headingColour),
            TextElement(self.app, 'Enable Sound', font,
                ScaledLocation(400, 495, 'topright'), colours.headingColour),
        ]

        initVolume = app.soundSettings.musicVolume
        musicVolumeLabel = TextElement(self.app, '%d' % (initVolume,), font,
                                 ScaledLocation(870, 280, 'topleft'),
                                 colours.headingColour)


        self.musicVolumeSlider = Slider(self.app,
                ScaledArea(450, 280, 400, 40))
        onSlide = lambda volume: musicVolumeLabel.setText("%d" % volume)
        self.musicVolumeSlider.onSlide.addListener(onSlide)
        self.musicVolumeSlider.onValueChanged.addListener(onSlide)
        self.musicVolumeSlider.setVal(initVolume)

        self.musicBox = CheckBox(self.app, ScaledLocation(450, 360),
                text = '', font = font, colour = (192,192,192),
                initValue = app.soundSettings.musicEnabled)

        initSndVolume = app.soundSettings.soundVolume
        soundVolumeLabel = TextElement(self.app, '%d' % (initSndVolume,), font,
                ScaledLocation(870, 420, 'topleft'), colours.headingColour)


        self.soundVolumeSlider = Slider(self.app,
                ScaledArea(450, 420, 400, 40))
        onSlide = lambda volume: soundVolumeLabel.setText("%d" % volume)
        self.soundVolumeSlider.onSlide.addListener(onSlide)
        self.soundVolumeSlider.onValueChanged.addListener(onSlide)
        self.soundVolumeSlider.setVal(initSndVolume)

        self.soundBox = CheckBox(self.app, ScaledLocation(450, 500),
                text = '', font = font, colour = (192,192,192),
                initValue = app.soundSettings.soundEnabled)

        self.buttons = [
            button(app, 'save', self.saveSettings, (-100, -75), 'midbottom',
                   secondColour=app.theme.colours.white),
            button(app, 'cancel', self.onClose.execute, (100, -75),
                'midbottom', secondColour=app.theme.colours.white)
        ]

        self.elements = text + [musicVolumeLabel, self.musicVolumeSlider,
                self.musicBox, soundVolumeLabel, self.soundVolumeSlider,
                self.soundBox] + self.buttons

    def saveSettings(self, sender=None):
        playMusic, volume, playSound, sndVolume = self.getValues()

        ss = self.app.soundSettings
        ss.musicEnabled = playMusic
        ss.musicVolume = volume
        ss.soundEnabled = playSound
        ss.soundVolume = sndVolume

        ss.save()
        ss.apply()

        self.onClose.execute()

    def getValues(self):

        playMusic = self.musicBox.value
        volume = self.musicVolumeSlider.getVal()

        playSound = self.soundBox.value
        sndVolume = self.soundVolumeSlider.getVal()

        return [playMusic, volume, playSound, sndVolume]




import logging
import os
import tempfile

from trosnoth.djangoapp.models import TrosnothServerSettings

log = logging.getLogger(__name__)


state = 'uninitialised'
murmur = None
createdRooms = []


def init(force=False):
    global state, murmur
    if state != 'uninitialised' and not force:
        return 'previously failed'

    try:
        settings = TrosnothServerSettings.get()
        if not settings.iceEnabled:
            state = 'disabled'
            return

        import Ice, IcePy

        proxyString = settings.iceProxyStringOverride
        if not proxyString:
            proxyString = 'Meta:tcp -h {host} -p {port} -t 1000'.format(
                host=settings.iceHost, port=settings.icePort)

        props = Ice.createProperties()
        props.setProperty('Ice.ImplicitContext', 'Shared')
        initData = Ice.InitializationData()
        initData.properties = props
        ice = Ice.initialize(initData)

        proxy = ice.stringToProxy(proxyString)
        op = IcePy.Operation(
            'getSlice', Ice.OperationMode.Idempotent,
            Ice.OperationMode.Idempotent, True, None, (), (), (),
            ((), IcePy._t_string, False, 0), ())

        sl = op.invoke(proxy, ((), None))
        fd, path  = tempfile.mkstemp(suffix='.ice')
        f = os.fdopen(fd, 'w')
        with f:
            f.write(sl)
            f.flush()

            Ice.loadSlice('', ['-I' + Ice.getSliceDir(), path])
        os.unlink(path)

        if settings.iceSecret:
            ice.getImplicitContext().put(
                'secret', settings.iceSecret.encode('ascii'))

        import Murmur
        murmur = Murmur.MetaPrx.checkedCast(proxy)
    except Exception:
        log.exception('Error initialising murmur connection')
        state = 'failed'
    else:
        state = 'initialised'
    return state


def setupRooms(teams, players):
    if state != 'initialised':
        return

    tearDownRooms()
    s = getServer()
    if not s:
        return

    teamToChannelId = {}
    for team in teams:
        channelId = s.addChannel(team.teamName, 0)
        createdRooms.append(channelId)
        teamToChannelId[team] = channelId

    users = {}
    for userState in s.getUsers().itervalues():
        users[userState.name.lower()] = userState

    for player in players:
        # Move the player into the appropriate channel
        if player.user is None:
            continue
        if player.team not in teams:
            continue
        key = player.user.username.lower()
        if key not in users:
            key = player.nick.lower()
            if key not in users:
                continue

        # Don't move this player again, even if it matches another player's
        # nick / username.
        userState = users.pop(key)

        userState.channel = teamToChannelId[player.team]
        s.setState(userState)


def tearDownRooms():
    if state != 'initialised':
        return
    s = getServer()
    if not s:
        return
    while createdRooms:
        roomId = createdRooms.pop()
        s.removeChannel(roomId)


def getServer():
    servers = murmur.getBootedServers()
    if servers:
        return servers[0]
    log.warning('Could not connect to murmur server via ICE')
    return None



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import pygame

from trosnoth.utils.event import Event
from trosnoth.gui.common import uniqueColour, Location, AttachedPoint

from trosnoth.gui.framework.framework import Element
from trosnoth.gui.framework.elements import TextButton
from trosnoth.gui.framework.utils import wrapWords

class ScrollingText(Element):
    def __init__(self, app, area, text, colour, fonts, bgColour=None,
            textAlign='left', loop=False, startOff=False, antiAlias=False):
        super(ScrollingText,self).__init__(app)

        self.area = area
        self.border = False
        self.shadowColour = None
        self.loop = loop
        self.startOff = startOff
        self.bgColour = bgColour
        self.speed = 50

        self.offset = 0
        self.colour = colour
        self.fonts = fonts
        self.text = text
        self.textAlign = textAlign
        self.antiAlias = antiAlias
        self.dirty = False

        # A test to see whether the size has changed
        self.lastSize = app.screenManager.size

        self._setImage()

        self.autoScroll = False


        # Create up and down buttons for if there's autoscroll.
        if fonts.has_key('buttons'):
            font = fonts['buttons']
        else:
            font = app.fonts.scrollingButtonsFont
        pos = Location(AttachedPoint((0,0), self._getRect, 'topright'),
                'topleft')
        self.upBtn = TextButton(app, pos, ' up', font, colour, (255,255,255))
        pos = Location(AttachedPoint((0,0), self._getRect, 'bottomright'),
                'bottomleft')
        self.dnBtn = TextButton(app, pos, ' down', font, colour, (255,255,255))

        self.onFinishedScrolling = Event()

    def setColour(self, colour):
        self.colour = colour
        self.dirty = True
        pos = Location(AttachedPoint((0,0), self._getRect, 'topright'),
                'topleft')
        if self.fonts.has_key('buttons'):
            font = self.fonts['buttons']
        else:
            font = self.app.fonts.scrollingButtonsFont
        self.upBtn = TextButton(self.app, pos, ' up', font, colour,
                (255,255,255))
        self.dnBtn = TextButton(self.app, pos, ' down', font, colour,
                (255,255,255))

    def setAutoscroll(self, hasAutoscroll):
        self.autoScroll = hasAutoscroll

    def setSpeed(self, speed):
        self.speed = speed

    def setBorder(self, hasBorder):
        self.border = hasBorder

    def setShadowColour(self, shadowColour):
        self.shadowColour = shadowColour
        self.dirty = True

    def _setImage(self):
        mainImage = self._setImageDetails(self.text, self.colour, self.fonts,
                self.textAlign)
        if self.shadowColour is None:
            self.image = mainImage
        else:
            shadowImage = self._setImageDetails(self.text, self.shadowColour,
                    self.fonts, self.textAlign)
            shadowOffset = (1, 1)
            size = mainImage.get_size()
            self.image = pygame.Surface(tuple([size[i] + shadowOffset[i] for i
                    in 0,1]))
            if self.bgColour:
                self.image.fill(self.bgColour)
            else:
                key = uniqueColour((self.colour, self.shadowColour))
                self.image.fill(key)
                self.image.set_colorkey(key)
            self.image.blit(shadowImage, shadowOffset)
            self.image.blit(mainImage, (0,0))
        self.dirty = False

    def _setImageDetails(self, text, colour, fonts, align):
        margin = 50

        lines = text.split('\n')
        rendered = []
        heading1Font = fonts['h1']
        heading2Font = fonts['h2']
        bodyFont = fonts['body']
        x = 0

        width = self._getSize()[0] - 2 * margin
        while x < len(lines):
            line = lines[x]
            if line.startswith("<h1>") and line.endswith("</h1>"):
                style = "h1"
                line = line[4:len(line)-5]
                font = heading1Font
            elif line.startswith("<h2>") and line.endswith("</h2>"):
                style = "h2"
                line = line[4:len(line)-5]
                font = heading2Font
            else:
                style = "body"
                font = bodyFont

            newLines = wrapWords(self.app, line, font, width)
            line = newLines[0]
            del newLines[0]

            # Insert the new lines into the list of lines
            newLines.reverse()
            for l in newLines:
                if style == "h1":
                    l = "<h1>" + l + "</h1>"
                elif style == "h2":
                    l = "<h2>" + l + "</h2>"
                lines.insert(x + 1, l)


            if self.bgColour:
                img = font.render(self.app, line, self.antiAlias, colour,
                        self.bgColour)
            else:
                img = font.render(self.app, line, self.antiAlias, colour)
            rendered.append(img)
            x += 1

        height = 0
        width = self._getSize()[0]
        for r in rendered:
            height += r.get_height()

        newImage = pygame.Surface((width, height))
        if self.bgColour:
            newImage.fill(self.bgColour)
        else:
            key = uniqueColour((colour),)
            newImage.fill(key)
            newImage.set_colorkey(key)
        yPos = 0
        for r in rendered:
            if align == 'left':
                xPos = margin
            elif align == 'right':
                xPos = width - r.get_width() - margin
            elif align == 'middle':
                xPos = (width - r.get_width()) / 2
            else:
                raise ValueError, "Not a valid alignment argument"

            newImage.blit(r, (xPos, yPos))
            yPos += r.get_height()
        if self.startOff:
            if self.loop:
                size = (width, height + self._getSize()[1])
            else:
                size = (width, height + 2 * self._getSize()[1])
        else:
            size = (width, height)

        image = pygame.Surface(size)
        if not self.bgColour:
            image.fill(key)
            image.set_colorkey(key)
        if self.startOff:
            image.blit(newImage, (0, self._getSize()[1]))
        else:
            image.blit(newImage, (0, 0))

        self.canScroll = image.get_height() > self._getSize()[1]
        self.reachedEnd = (not self.loop and self.offset == 0) or (self.loop
                and not self.canScroll)

        return image

    def returnToTop(self):
        self.offset = 0

    def _getRect(self):
        return self.area.getRect(self.app)

    def _getSize(self):
        return self._getRect().size

    def _getPt(self):
        return self._getRect().topleft

    def draw(self, screen):
        if self.dirty or self.app.screenManager.size != self.lastSize:
            self.lastSize = self.app.screenManager.size
            self._setImage()
        if not self.canScroll:
            # Our image won't cover the whole of the scrollingText
            height = self._getSize()[1] - self.image.get_height()
            rect = pygame.Rect(self._getPt()[0], self._getPt()[1] +
                    self.image.get_height(), self._getSize()[0], height)
            if self.bgColour:
                screen.fill(self.bgColour, rect)
        # Find the segment which will be drawn on screen
        rect = pygame.Rect((0, self.offset), self._getSize())
        screen.blit(self.image, self._getPt(), rect)

        if (self.loop and self.canScroll and self.offset + self._getSize()[1] >
                self.image.get_height()):
            # It's doing the looping
            rect = pygame.Rect((0, self.offset - self.image.get_height()),
                    self._getSize())
            screen.blit(self.image, self._getPt(), rect)

        if self.border:
            rect.topleft = self._getPt()
            pygame.draw.rect(screen, (0,0,0), rect, 4)

        if not self.autoScroll:
            self.upBtn.draw(screen)
            self.dnBtn.draw(screen)

    def processEvent(self, event):
        if self.autoScroll:
            # We don't consume any events.
            return event
        else:
            event = self.upBtn.processEvent(event)
            if not event: return
            event = self.dnBtn.processEvent(event)
            if not event: return

            return event

    def scroll(self, offset):
        if not self.canScroll:
            return
        if self.reachedEnd:
            if ((offset > 0 and self.offset > 0) or
                    (offset < 0 and self.offset == 0)):
               # They have already reached the end, and are trying to
               # go further that way - do nothing.
               return
        self.reachedEnd = False
        newOffset = offset + self.offset
        if not self.loop:
            if newOffset + self._getSize()[1] > self.image.get_height():
                # They've hit the bottom
                newOffset = self.image.get_height() - self._getSize()[1]
                self.onFinishedScrolling.execute()
                self.reachedEnd = True
            elif newOffset <= 0:
                # Hit the top
                newOffset = 0
                self.onFinishedScrolling.execute()
                self.reachedEnd = True
        self.offset = newOffset

        # For looping: make sure it loops around
        if self.loop:
            self.offset %= self.image.get_height()


    def _getSpeed(self):
        if hasattr(self.speed, 'getSpeed'):
            speed = self.speed.getSpeed(self.app)
        else:
            speed = self.speed
        return speed

    def tick(self, deltaT):
        if self.autoScroll:
            self.scroll(self._getSpeed() * deltaT)
        else:
            if self.upBtn.mouseOver:
                self.scroll(-self._getSpeed() * deltaT)
            elif self.dnBtn.mouseOver:
                self.scroll(self._getSpeed() * deltaT)
            self.upBtn.tick(deltaT)
            self.dnBtn.tick(deltaT)



import pygame

from trosnoth.trosnothgui.settings.display import DisplaySettingsTab
from trosnoth.trosnothgui.settings.keymap import KeymapTab
from trosnoth.trosnothgui.settings.sound import SoundSettingsTab
from trosnoth.trosnothgui.settings.themes import ThemeTab
from trosnoth.gui.framework import framework, elements
from trosnoth.gui.framework.tabContainer import TabContainer, TabSize
from trosnoth.gui.common import (ScaledArea, Location, AttachedPoint)
from trosnoth.utils.event import Event

class SettingsMenu(framework.CompoundElement):
    def __init__(self, app, onClose=None, onRestart=None,
            showThemes=False, showSound=True, showDisplay=True,
            showKeymap=True):
        super(SettingsMenu, self).__init__(app)

        self.onClose = Event()
        if onClose is not None:
            self.onClose.addListener(onClose)
        self.onRestart = Event()
        if onRestart is not None:
            self.onRestart.addListener(onRestart)

        area = ScaledArea(50,140,924, 570)
        bg = pygame.Surface((924, 500))
        bg.fill(app.theme.colours.settingsMenu)
        if app.displaySettings.alphaOverlays:
            bg.set_alpha(192)
        font = app.screenManager.fonts.bigMenuFont
        self.tabContainer = TabContainer(self.app, area, font,
                app.theme.colours.settingsTabBorder)
        bp = elements.SizedPicture(app, bg, Location(AttachedPoint((0,0),
                self.tabContainer._getTabRect)), TabSize(self.tabContainer))

        if showDisplay:
            displayTab = DisplaySettingsTab(app, onClose=self.onClose.execute)
            self.tabContainer.addTab(displayTab)

        if showKeymap:
            keymapTab = KeymapTab(app, onClose=self.onClose.execute)
            self.tabContainer.addTab(keymapTab)

        if showSound:
            soundTab = SoundSettingsTab(app, onClose=self.onClose.execute)
            self.tabContainer.addTab(soundTab)

        if showThemes:
            themeTab = ThemeTab(app, onClose=self.onClose.execute,
                    onRestart=self.onRestart.execute)
            self.tabContainer.addTab(themeTab)

        self.elements = [bp, self.tabContainer]



from trosnoth.gui.framework import framework, elements
from trosnoth.gui import colours as guiColours

from trosnoth.gui.common import RelativeLocation


class MessageBank(framework.CompoundElement):
    '''Defines the implementation for a queue of Messages to be displayed
    on screen'''

    def __init__(
            self, app, maxSize, maxLength, startPt,
            L_R, U_D, font, timeLimit=10):
        '''
        @param maxSize: The largest number of lines that can be displayed at
            any given time.
        @param maxLength: The maximum length of a line. Lines longer than this
            will be wrapped.
        @param startPt: The corner closest to the edge of the screen that a
            message will get. For example, if aligned top left, this point
            will specify the top left position of the most recent message.
        @param L_R: the alignment of the text that appears ("left" or "right")
        @param U_D: defines whether the text will be coming from the top or the
            bottom ("top" or "bottom")
        @param timeLimit: the time (in seconds) before a message disappears
        '''
        super(MessageBank, self).__init__(app)
        self.timeLimit = timeLimit
        if maxSize <= 0:
            raise ValueError('Must be able to display at least one message')
        # Stores all messages again, so that we have an order
        self.elements = []
        self.font = font
        self.maxSize = maxSize
        self.maxLength = maxLength
        self.currentLength = 0
        # Left or right-aligned; at top or bottom of screen.
        self.L_R = L_R
        self.U_D = U_D
        self.startPt = startPt

    def newMessage(self, text, colour):
        '''
        Adds a new message to the bank, splitting lines as necessary.

        @param text: The text of the message. May contain newline characters.
        @param colour: A tuple.
        '''
        text = self.breakUp(text)
        if not text.__contains__('\n'):
            self._newIndividual(Message(
                self.app, text, colour, self, self.timeLimit))
            return
        tempMessages = []
        i = 0
        while 1:
            while i < len(text) and text[i] != '\n':
                i += 1
            tempMessages.append(
                Message(self.app, text[:i], colour, self, self.timeLimit))
            text = text[i+1:]
            if i == len(text):
                break
            i = 0

        if self.U_D == "bottom":
            inc = 1
            j = 0
        else:
            inc = -1
            j = len(tempMessages) - 1

        while j >= 0 and j < len(tempMessages):
            self._newIndividual(tempMessages[j])
            j += inc

    def _newIndividual(self, newMessage):
        self.prepareForInsertion()
        self.elements.append(newMessage)

    def prepareForInsertion(self):
        '''
        Makes room for a new message.

        Deletes the oldest message if the capacity has been reached and moves
        every message up one space.
        '''
        if self.currentLength >= self.maxSize:
            self.delMessage(True)
        for message in self.elements:
            message.downByOne()
        self.currentLength += 1

    def newColourMessage(self, parts):
        '''
        Adds a message with multiple colour parts to the MessageBank.
        This method does not split / wrap lines: you have to do this yourself!

        @param parts: A list containing the colour segments of the message.
        Each segment is a tuple with the text and colour for that segment.
        '''
        self.prepareForInsertion()

        count = 0
        xOffset = 0

        for text, colour in parts:
            count += 1

            # Only the last item should have a break afterwards
            if count == len(parts):
                trailingBreak = True
            else:
                trailingBreak = False

            message = Message(
                self.app, text, colour, self,
                self.timeLimit, trailingBreak, xOffset)

            xOffset += message._getRect().width

            self.elements.append(message)

    def delMessage(self, popped=False):
        # We'll always be deleting the oldest message
        popped = self.elements.pop(0)
        while not popped.trailingBreak:
            popped = self.elements.pop(0)
        self.currentLength -= 1

    def breakUp(self, text):
        '''Breaks the text up into separate lines of max 50 chars. If there's
        a space within 20 characters of the end, it will break there, otherwise
        it will simply cut the current word in half'''
        maxLength = self.maxLength
        minLength = 30
        lineDelimiters = (' ', '\t', '-')
        i = 0
        pos = 0
        modPos = 0
        while i < len(text):
            char = text[i]
            # Line Delimiter found; make a record of it
            if char in lineDelimiters:
                pos = i + 1
            modPos += 1
            # Gone too far; break the line
            if modPos >= maxLength:

                # No line delimiters close to the end of this line; break the
                # current word
                if pos < minLength:
                    text = text[:i] + '-\n' + text[i:]
                    i += 2
                # Break at the latest line delimiter
                else:
                    text = text[:pos - 1] + '\n' + text[pos:]
                    # Since a new line has begun back a bit, set it back
                    i = pos
                modPos = 0
                pos = 0
            i += 1
        # Remove trailing whitespace
        while text.endswith('\n') or text.endswith(' '):
            text = text[:len(text) - 1]
        return text

    def deleteAll(self):
        while len(self.elements) > 0:
            self.delMessage(True)


class Message(elements.TextElement):
    def __init__(
            self, app, text, colour, messageBank, timeLimit,
            trailingBreak=True, xOffset=0):
        self.messageBank = messageBank
        if not colour:
            colour = app.theme.colours.grey

        super(Message, self).__init__(
            app, text, self.messageBank.font, self.messageBank.
            startPt, colour, shadow=True)

        if xOffset > 0:
            self.setPos(RelativeLocation(self.pos, (xOffset, 0)))

        if colour == guiColours.shadowDefault:
            self.setShadowColour(guiColours.shadowAgainstGrey)

        self.timeLeft = timeLimit
        self.xOffset = xOffset
        self.trailingBreak = trailingBreak

    def __repr__(self):
        return 'Message object: "%s"' % self.getText()

    def downByOne(self):
        # Defines where on the screen it is displayed
        if self.messageBank.U_D == "top":
            pos = RelativeLocation(
                self.pos, (0, self.image.font.getHeight(self.app)))
        else:
            pos = RelativeLocation(
                self.pos, (0, -self.image.font.getHeight(self.app)))
        self.setPos(pos)

    def tick(self, deltaT):
        if self.timeLeft is not None:
            self.timeLeft -= deltaT
            if self.timeLeft <= 0:
                self.messageBank.delMessage()



import logging

import pygame

from trosnoth.gui.common import Location
from trosnoth.gui.framework import elements, framework
from trosnoth.utils.utils import wrapline

log = logging.getLogger(__name__)


class CollapseBox(framework.CompoundElement):
    ACCELERATION = 500  # pix/s/s

    def __init__(
            self, app, region, titleFont, font, titleColour, hvrColour, colour,
            backColour, title='Note', lines=()):
        super(CollapseBox, self).__init__(app)

        self.region = region
        self.titleFont = titleFont
        self.font = font
        self.titleColour = titleColour
        self.hvrColour = hvrColour
        self.colour = colour
        self.backColour = backColour
        self.title = title
        self.lines = lines
        self.rect = None
        self.targetRect = None
        self.background = None
        self.hidden = False
        self.velocity = 0
        self.updateElements()

        app.screenManager.onResize.addListener(self.appResized)

    def appResized(self):
        self.updateElements()

    def setInfo(self, lines, title=None):
        if lines == self.lines and (title is None or title == self.title):
            return

        if title is not None:
            self.title = title
        self.lines = lines

        self.updateElements()

    def recalculateTarget(self):
        targetRect = self.region.getRect(self.app)
        titleBtn = self.makeTitleButton(targetRect)
        results = [titleBtn]

        y = targetRect.top + titleBtn.stdImage.get_height() + 4
        maxWidth = targetRect.width - 4
        font = self.font._getFont(self.app)
        if not self.hidden:
            x = targetRect.left + 2
            for line in self.lines:
                for subline in wrapline(line, font, maxWidth):
                    element = elements.TextElement(
                        app=self.app, text=subline, font=self.font,
                        pos=Location((x, y), 'topleft'),
                        colour=self.colour)
                    results.append(element)
                    y += element.image.getImage(self.app).get_height() + 2
        targetRect.height = y - targetRect.top
        return targetRect, results

    def updateElements(self):
        self.targetRect, elements = self.recalculateTarget()
        if self.rect is None:
            self.rect = self.targetRect
        else:
            self.rect.topleft = self.targetRect.topleft
            self.rect.width = self.targetRect.width

        self.background = pygame.Surface(self.rect.size, pygame.SRCALPHA)
        self.background.fill(self.backColour)

        if self.hidden or self.rect.size != self.targetRect.size:
            elements[1:] = []

        self.elements = elements

    def makeTitleButton(self, rect):
        if self.hidden:
            text = u'\u25b8 ' + self.title
        else:
            text = u'\u25be ' + self.title
        img = self.titleFont.render(self.app, text, True, self.titleColour)
        hvr = self.titleFont.render(self.app, text, True, self.hvrColour)

        btn = elements.HoverButton(
            app=self.app,
            pos=Location((rect.left + 2, rect.top + 2), 'topleft'),
            stdImage=img, hvrImage=hvr,
            onClick=self.toggle,
        )
        return btn

    def draw(self, screen):
        if not (self.title or self.lines):
            return
        screen.blit(self.background, self.rect)
        super(CollapseBox, self).draw(screen)

    def toggle(self, element):
        self.hidden = not self.hidden
        self.updateElements()

    def tick(self, deltaT):
        super(CollapseBox, self).tick(deltaT)

        diff = self.targetRect.height - self.rect.height
        if diff == 0:
            return

        sgn = diff / abs(diff)
        diff = abs(diff)
        self.velocity = sgn * min(
            abs(self.velocity) + deltaT * self.ACCELERATION,
            (2.0 * self.ACCELERATION * diff) ** 0.5)
        self.rect.height += sgn * min(abs(self.velocity) * deltaT + 1, diff)
        self.updateElements()



from trosnoth.ais.alpha import AlphaAI

from twisted.internet import task


class TestingAI(AlphaAI):
    nick = 'TestBot'
    playable = True

    # Use these toggles to change the behaviour of the bots.
    canMove = True
    canShoot = True
    canJump = True
    upgradeType = None

    def start(self):
        super(TestingAI, self).start()
        self._loop = task.LoopingCall(self.tick)
        self._loop.start(0.5)

    def tick(self):
        if (not self.player.dead and self.player.upgrade is None
               and self.upgradeType is not None):
            self.doBuyUpgrade(self.upgradeType)

    def shoot(self):
        if self.canShoot:
            super(TestingAI, self).shoot()

    def changeXDir(self):
        if self.canMove:
            super(TestingAI, self).changeXDir()

    def jumpAgain(self):
        if self.canJump:
            super(TestingAI, self).jumpAgain()

AIClass = TestingAI



'''
The hub/node interface exists primarily for sending and receiving network
messages, but it is also used when Trosnoth is run in isolated mode (for
testing).
This interface provides a channel for messages to be passed between agents and
a game. A hub sits on the game side of this interface, while a node sits on the
agent side. The interface may be used to pass messages for zero or more agents.
'''

MAX_AGENT_ID = 1 << 16 - 1


class Hub(object):
    '''
    Sits on the game side of the hub/node message passing interface.
    '''

    def __init__(self, *args, **kwargs):
        super(Hub, self).__init__(*args, **kwargs)
        self.node = None

    def connectNode(self, node):
        assert self.node is None
        assert node.hub is None
        self.node = node
        node.hub = self

    def disconnectNode(self):
        self.node.hub = None
        self.node = None

    def connectNewAgent(self):
        '''
        Requests that a new agent be connected to the game that this hub
        represents. If successful, this method will return an agentId for the
        new agent (or a deferred that results in the agentId).
        May raise UnableToConnect.
        '''
        raise NotImplementedError(
            '%s.connectNewAgent' % (self.__class__.__name__,))

    def disconnectAgent(self, agentId):
        '''
        Requests that the agent with the given id be disconnected from the
        game. May return a deferred that will not complete until the agent is
        disconnected.
        Note that the onAgentDisconnected event will fire whenever an agent is
        disconnected, regardless of whether it is triggered by this function or
        some other reason (e.g. lost network connection).
        '''
        raise NotImplementedError(
            '%s.disconnectAgent' % (self.__class__.__name__,))

    def sendRequestToGame(self, agentId, msg):
        raise NotImplementedError(
            '%s.sendRequestToGame' % (self.__class__.__name__,))


class UnableToConnect(Exception):
    '''
    Raised by connectNewAgent if the connection could not be established.
    '''


class Node(object):
    '''
    Sits on the agent side of the hub/node message passing interface.
    '''

    def __init__(self, *args, **kwargs):
        super(Node, self).__init__(*args, **kwargs)
        self.hub = None

    def gotServerCommand(self, msg):
        '''
        Called by the connected hub for every message from the server.
        '''
        raise NotImplementedError(
            '%s.gotServerCommand' % (self.__class__.__name__,))

    def gotMessageToAgent(self, agentId, msg):
        '''
        Called when the game sends a message targetted at a specific agent.
        '''
        raise NotImplementedError(
            '%s.gotMessageToAgent' % (self.__class__.__name__,))

    def agentDisconnected(self, agentId):
        '''
        Called when one of this node's agents has been disconnected from the
        game for any reason.
        '''
        raise NotImplementedError(
            '%s.agentDisconnected' % (self.__class__.__name__,))



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import pygame

from trosnoth.gui.app import Application
from trosnoth.gui.framework.framework import CompoundElement
from trosnoth.gui.framework.elements import SizedPicture
from trosnoth.gui.common import Location
from trosnoth.utils import datafile

from trosnoth.gui.framework.menu import MenuDisplay
from trosnoth.gui.menu.menu import MenuManager, Menu, MenuItem

class Interface(CompoundElement):
    backdropLocator = datafile.Locator('startupMenu')

    def __init__(self, app):
        super(Interface, self).__init__(app)
        
        self.menu = None
        self.elements = []
        self.makeElements()

    def makeElements(self):
        backdrop = pygame.image.load(self.backdropLocator \
                                     .getPath('backdrop.png')).convert()
        backdrop = SizedPicture(self.app, backdrop, Location((0,0), 'topleft'),
                                 FullScreenSize())

        menuMan = MenuManager()
        menu1 = Menu(name='File',
                     items=[MenuItem('New'),
                            MenuItem('Open'),
                            MenuItem('Save'),
                            MenuItem('Save As'),
                            MenuItem('Exit', self.app.stop),
                            MenuItem('---'),
                            MenuItem('Cancel', menuMan.cancel)]
                     )
        menu0 = Menu(name='Main',
                     items=[MenuItem('File', lambda: menuMan.showMenu(menu1)),
                            MenuItem('Blah')]
                     )
        menuMan.setDefaultMenu(menu0)
        
        self.menu = MenuDisplay(self.app,
                                location=Location((300, 400), 'bottomleft'),
                                size=(200, 200),
                                font=pygame.font.Font(None, 28),
                                manager=menuMan, titleColour=(255, 255, 0),
                                stdColour=(255, 255, 128),
                                hvrColour=(128, 255, 255),
                                backColour=(0, 64, 192), autosize=True,
                                hidable=True)
        
        self.elements = [backdrop, self.menu]



if __name__ == '__main__':
    screenSize = (920, 690)
    graphicsOptions = 0
    caption = 'Trosnoth Menu Test Program'
    
    app = Application(screenSize, graphicsOptions, caption, Interface)
    app.run()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging
import struct

from twisted.internet import reactor

from trosnoth.const import TICK_PERIOD
from trosnoth.messages import TickMsg
from trosnoth.model.hub import Hub
from trosnoth.network.networkDefines import serverVersion
from trosnoth.network.client import clientMsgs
from trosnoth.utils.netmsg import MessageTypeError
from trosnoth.utils.twist import WeakLoopingCall
from trosnoth.utils.unrepr import unrepr

log = logging.getLogger(__name__)


class ReplayRecorder(object):
    def __init__(self, world, filename):
        self.filename = filename
        self.world = world
        self.file = open(self.filename, 'wb')
        self.timeStarted = self.world.timer.getElapsedTime()
        self.stopped = False

        initialData = self.world.dumpEverything()
        initialData['serverVersion'] = serverVersion
        data = repr(initialData)
        self.file.write(struct.pack('!I', len(data)))
        self.file.write(data)

    def consumeMsg(self, msg):
        if self.stopped:
            return
        msgData = msg.pack()
        self.file.write(struct.pack('!I', len(msgData)))
        self.file.write(msgData)

    def stop(self):
        if not self.stopped:
            self.file.close()
            self.stopped = True


class ReplayPlayer(Hub):
    '''
    Emulates a normal server by outputting the same messages that a server once
    did.
    '''

    CHUNK_SIZE = 8000

    def __init__(self, filename, *args, **kwargs):
        super(ReplayPlayer, self).__init__(*args, **kwargs)
        self.tickPeriod = TICK_PERIOD
        self.file = open(filename, 'rb')
        self.finished = False
        self.loop = WeakLoopingCall(self, 'tick')
        self.agentIds = []
        self.nextAgentId = 0

        length = struct.unpack('!I', self.file.read(4))[0]
        data = self.file.read(length)
        self.settings = unrepr(data)

    def popSettings(self):
        result, self.settings = self.settings, None
        return result

    def start(self):
        self.loop.start(self.tickPeriod)

    def tick(self):
        while True:
            data = self.file.read(4)
            if not data:
                self.finished = True
                if self.node:
                    while self.agentIds:
                        agentId = self.agentIds.pop(0)
                        # Give 2 seconds before ending the replay
                        reactor.callLater(
                            2, self.node.agentDisconnected, agentId)
                self.loop.stop()
                break
            length = struct.unpack('!I', data)[0]

            data = self.file.read(length)
            try:
                msg = clientMsgs.buildMessage(data)
            except MessageTypeError:
                log.warning('WARNING: UNKNOWN MESSAGE: %r' % (data,))
                continue

            self.node.gotServerCommand(msg)
            if isinstance(msg, TickMsg):
                break

    def connectNewAgent(self):
        agentId = self.nextAgentId
        self.nextAgentId += 1
        self.agentIds.append(agentId)
        return agentId

    def disconnectAgent(self, agentId):
        self.agentIds.remove(agentId)

    def sendRequestToGame(self, agentId, msg):
        pass



from trosnoth.gui.framework import framework
from trosnoth.gui.framework.elements import TextElement, Backdrop
from trosnoth.version import titleVersion

from trosnoth.gui.common import (Location, ScaledSize,
        FullScreenAttachedPoint)

class TrosnothBackdrop(framework.CompoundElement):
    def __init__(self, app):
        super(TrosnothBackdrop, self).__init__(app)

        backdropPath = app.theme.getPath('startupMenu', 'blackdrop.png')
        backdrop = Backdrop(app, backdropPath,
                app.theme.colours.backgroundFiller)

        # Things that will be part of the backdrop of the entire startup menu
        # system.
        verFont = self.app.screenManager.fonts.versionFont
        self.elements = [
            backdrop,
            TextElement(self.app, titleVersion, verFont,
                Location(FullScreenAttachedPoint(ScaledSize(-10,-10),
                'bottomright'), 'bottomright'),
                app.theme.colours.versionColour),
        ]




# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

'''
Originally based on the Python code module.
'''

from code import InteractiveConsole
import pygame
import sys
from twisted.internet import defer
from trosnoth.gui.framework import framework
from trosnoth.gui.framework.prompt import InputBox
from trosnoth.gui.common import ScaledArea

def softspace(file, newvalue):
    oldvalue = 0
    try:
        oldvalue = file.softspace
    except AttributeError:
        pass
    try:
        file.softspace = newvalue
    except (AttributeError, TypeError):
        # "attribute-less object" or "read-only attributes"
        pass
    return oldvalue

class TwistedInteractiveConsole(InteractiveConsole):
    """Closely emulate the behavior of the interactive Python interpreter.

    This class builds on InteractiveInterpreter and adds prompting
    using the familiar sys.ps1 and sys.ps2, and input buffering.

    """

    def __init__(self, *args, **kwargs):
        InteractiveConsole.__init__(self, *args, **kwargs)
        self._dummyStdOut = DummyStdOut(self.write)

    @defer.inlineCallbacks
    def interact(self, banner=None):
        """Closely emulate the interactive Python console.

        The optional banner argument specify the banner to print
        before the first interaction; by default it prints a banner
        similar to the one printed by the real Python interpreter,
        followed by the current class name in parentheses (so as not
        to confuse this with the real interpreter -- since it's so
        close!).

        """
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "
        cprt = ('Type "help", "copyright", "credits" or "license" for more '
                'information.')
        if banner is None:
            self.write("Python %s on %s\n%s\n(%s)\n" %
                       (sys.version, sys.platform, cprt,
                        self.__class__.__name__))
        else:
            self.write("%s\n" % str(banner))
        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2
                else:
                    prompt = sys.ps1
                try:
                    line = yield self.raw_input(prompt)

                    # Can be None if sys.stdin was redefined
                    encoding = getattr(sys.stdin, "encoding", None)
                    if encoding and not isinstance(line, unicode):
                        line = line.decode(encoding)
                except EOFError:
                    self.write("\n")
                    break
                else:
                    try:
                        more = self.push(line)
                    except SystemExit:
                        break
            except KeyboardInterrupt:
                self.write("\nKeyboardInterrupt\n")
                self.resetbuffer()
                more = 0

    def raw_input(self, prompt=""):
        """Write a prompt and read a line.

        The returned line does not include the trailing newline.
        When the user enters the EOF key sequence, EOFError is raised.

        The base implementation uses the built-in function
        raw_input(); a subclass may replace this with a different
        implementation.

        """
        return defer.succeed(raw_input(prompt))

    def runcode(self, code):
        stdout = sys.stdout
        sys.stdout = self._dummyStdOut
        try:
            return InteractiveConsole.runcode(self, code)
        finally:
            sys.stdout = stdout

class DummyStdOut(object):
    def __init__(self, write):
        self.write = write
    def flush(self):
        pass

class DummyStdIn(object):
    def __init__(self, read):
        self.read = read

    def readline(self):
        result = ''
        while True:
            c = self.read(1)
            result += c
            if result in ('', '\n'):
                return result

class TrosnothInteractiveConsole(TwistedInteractiveConsole,
        framework.Element):

    def __init__(self, app, font, area, colour=(192,192,192),
            backColour=(0, 64, 0), historyLimit=2000,
            locals={}):
        TwistedInteractiveConsole.__init__(self, locals=locals)
        framework.Element.__init__(self, app)

        self.font = font
        self.area = area
        self.colour = colour
        self.backColour = backColour
        self.rect = None
        self.image = None
        self.dirty = True
        self.history = ['']
        self.historyLimit = historyLimit
        self.inBox = ib = InputBox(self.app, ScaledArea(0, 0, 1000, 100),
                font=self.font)
        ib.onEnter.addListener(self._enter)
        ib.hasFocus = True
        ib.gotFocus()

        self.readDeferred = None
        self.deferredInputs = []   # List of (prompt, deferred)
        self.font = font
        self.typeAhead = []
        self.inText = ''
        self.cursorPos = 0
        self._dummyStdIn = DummyStdIn(self.stdInRead)
        self._dummyStdIn.deferred_raw_input = self.raw_input

    def raw_input(self, prompt=""):
        """Write a prompt and read a line.

        The returned line does not include the trailing newline.
        When the user enters the EOF key sequence, EOFError is raised.

        The base implementation uses the built-in function
        raw_input(); a subclass may replace this with a different
        implementation.

        """
        if self.readDeferred is not None:
            d = defer.Deferred()
            self.deferredInputs.append((prompt, d))
            return d
        self.write(prompt)
        if len(self.typeAhead) > 0:
            val = self.typeAhead.pop(0)
            return defer.succeed(val)

        self.readDeferred = d = defer.Deferred()
        return d

    def _nextRead(self):
        '''
        Begins the next queued readline if there is one.
        '''
        if len(self.deferredInputs) == 0:
            self.readDeferred = None
            return

        prompt, d = self.deferredInputs.pop(0)
        self.write(prompt)
        self.readDeferred = d

    def _enter(self, box):
        val = box.getValue()
        self.write(val + '\n')
        if self.readDeferred is None:
            self.typeAhead.append(val)
        else:
            d = self.readDeferred
            self._nextRead()
            d.callback(val)
        self.inBox.setValue('')

    def write(self, data):
        data = self.history.pop(-1) + data
        self.history.extend(data.split('\n'))
        self.history = self.history[-self.historyLimit:]
        self.dirty = True

    def runcode(self, code):
        stdin = sys.stdin
        sys.stdin = self._dummyStdIn
        try:
            return TwistedInteractiveConsole.runcode(self, code)
        finally:
            sys.stdin = stdin

    def stdInRead(self, count=None):
        if count is None:
            # Read all the type ahead.
            result = '\n'.join(self.typeAhead) + '\n'
            self.typeAhead[:] = []
        else:
            result = ''
            while len(self.typeAhead) > 0:
                if count <= len(self.typeAhead[0]) + 1:
                    result += self.typeAhead[0][:count]
                    self.typeAhead[0] = self.typeAhead[0][count:]
                    break
                line = self.typeAhead.pop(0)
                count -= len(line) + 1
                result += line + '\n'
        return result

    def draw(self, screen):
        r = self.area.getRect(self.app)
        if r != self.rect:
            self.rect = r
            self.dirty = True
            self.image = pygame.Surface(self.rect.size).convert()
        if self.dirty:
            self._redraw()

        if (self.inText != self.inBox.getValue() or self.cursorPos !=
                self.inBox.cursorIndex):
            self._drawLastLine()

        screen.blit(self.image, self.rect.topleft)

    def processEvent(self, event):
        return self.inBox.processEvent(event)

    def _redraw(self):
        if self.backColour is None:
            self.image.fill((255, 255, 254))
            self.image.set_colorkey((255, 255, 254))
        else:
            self.image.fill(self.backColour)

        y = self.rect.height
        i = len(self.history) - 1
        wrappedLines = []
        while y > 0 and (i >= 0 or len(wrappedLines) > 0):
            if len(wrappedLines) == 0:
                wrappedLines = self._wrapLine(self.history[i])
                i -= 1

            text = wrappedLines.pop()
            if self.backColour is None:
                line = self.font.render(self.app, text, False, self.colour)
            else:
                line = self.font.render(self.app, text, True, self.colour,
                        self.backColour)
            y -= line.get_height()
            self.image.blit(line, (4, y))

        self.dirty = False

    def _wrapLine(self, text):
        '''
        Takes the given line of text, decides where it should be wrapped, and
        returns a list of new lines of text to write.
        '''
        result = []
        indent = 0
        while len(text) > indent:
            n = max(1, self._getLineLength(text))
            if n <= indent * 1.5 and indent > 0:
                text = text[indent:]
                n = max(1, self._getLineLength(text))

            result.append(text[:n])
            if indent == 0:
                indent = 8
                for c in text:
                    if c.isspace():
                        indent += 1
                    else:
                        break
            text = ' ' * indent + text[n:]
        if result == []:
            result = ['']
        return result

    def _getLineLength(self, text):
        '''
        Works out the number of characters into the given string at which the
        string should be wrapped.
        '''
        width = self.rect.width - 8
        result = big = len(text)
        small = 0

        # Binary search to find the largest value that fits in.
        while big != small:
            textWidth = self.font.size(self.app, text[:result])[0]
            if textWidth > width:
                big = result - 1
            else:
                small = result
            result = (big + small + 1) / 2

        result = big
        def doNotSplit(char):
            return char.isalnum() or char in '_\'"'

        if result == len(text) or not doNotSplit(text[result]):
            return result
        while result > 0 and doNotSplit(text[result-1]):
            result -= 1

        if result == 0:
            result = big

        return result

    def _drawLastLine(self):
        self.inText = inText = self.inBox.getValue()
        self.cursorPos = cPos = self.inBox.cursorIndex

        text = self.history[-1] + inText[:cPos] + '|' + inText[cPos:]
        if self.backColour is None:
            line = self.font.render(self.app, text, False, self.colour)
        else:
            line = self.font.render(self.app, text, True, self.colour,
                    self.backColour)
        fillRect = line.get_rect()
        fillRect.bottom = self.rect.height
        fillRect.width = self.rect.width
        if self.backColour is None:
            self.image.fill((255, 255, 254), fillRect)
        else:
            self.image.fill(self.backColour, fillRect)
        self.image.blit(line, (4, fillRect.top))




# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import pickle
from twisted.protocols import amp

class AuthServerError(Exception):
    pass

class NotAuthenticated(AuthServerError):
    pass

class PortUnreachable(AuthServerError):
    '''
    Raised by the server during an attempt to register a game if the specified
    port cannot be reached.
    '''

class CannotCreateGame(AuthServerError):
    '''
    Raised by the server during an attempt to create a game if the game is not
    allowed, perhaps because there are already the maximum number of games
    running on this server.
    '''

class GameDoesNotExist(AuthServerError):
    '''
    Cannot join the given game because it does not exist.
    '''

class ListGames(amp.Command):
    '''
    Returns a list of the Trosnoth games available on this server.
    '''
    arguments = []
    response = [
        ('games', amp.AmpList([
            ('id', amp.Integer()),
            ('game', amp.String()),
            ('version', amp.String()),
        ]))
    ]

class ListOtherGames(amp.Command):
    '''
    Returns a list of Trosnoth games this server knows about which are not
    hosted on this server.
    '''
    arguments = []
    response = [
        ('games', amp.AmpList([
            ('ip', amp.String()),
            ('port', amp.Integer()),
            ('game', amp.String()),
            ('version', amp.String()),
        ]))
    ]

class RegisterGame(amp.Command):
    '''
    Registers a game running on the client with the server.
    '''
    arguments = [
        ('game', amp.String()),
        ('version', amp.String()),
        ('port', amp.Integer())
    ]
    response = []
    errors = {
        NotAuthenticated: 'NO_AUTH',
        PortUnreachable: 'BAD_PORT',
    }

class CreateGame(amp.Command):
    '''
    Instructs the server to start a new game.
    '''
    arguments = [
        ('game', amp.String()),
    ]
    response = [
        ('id', amp.Integer()),
        ('version', amp.String()),
    ]
    errors = {
        NotAuthenticated: 'NO_AUTH',
        CannotCreateGame: 'WONT',
    }

class ConnectToGame(amp.Command):
    '''
    Requests details with which to join a particular game on the server.
    nick is the preferred nick of the authenticated player.
    '''
    arguments = [('id', amp.Integer())]
    response = [('port', amp.Integer()),
                ('authTag', amp.Integer()),
                ('nick', amp.Unicode())]
    errors = {
        NotAuthenticated: 'NO_AUTH',
        GameDoesNotExist: 'NO_GAME',
    }


class GetAuthToken(amp.Command):
    '''
    Requests the server send a random string to be used during authentication or
    creation of a user. The authentication token which is returned is valid
    until GetAuthToken is called again.
    '''
    arguments = []
    response = [('token', amp.String())]

class PasswordAuthenticate(amp.Command):
    '''
    Authenticates with the server using a password. The password value which is
    sent is created by prepending an authentication token obtained from
    GetAuthToken to the password, then encrypting using the server's public RSA
    key.
    '''
    arguments = [
        ('username', amp.String()),
        ('password', amp.String()),
    ]
    response = [('result', amp.Boolean())]

class SetPassword(amp.Command):
    arguments = [
        ('password', amp.String()),
    ]
    response = []
    errors = {
        NotAuthenticated: 'NO_AUTH',
    }

class GetSupportedSettings(amp.Command):
    '''
    Returns a list of strings which indicate which account settings the
    client should display to the user.
    Possible values are 'password'
    '''
    arguments = []
    response = [
        ('result', amp.ListOf(amp.String())),
    ]

class CreateUserWithPassword(amp.Command):
    '''
    Requests to create a user on the server using the given username and
    password. The password value which is sent is created by prepending an
    authentication token obtained from GetAuthToken to the password, then
    encrypting using the server's public RSA key.
    If successful, the result string will be empty, otherwise it will be a
    human-readable reason why the server is not allowing new users to be
    created.
    '''
    arguments = [
        ('username', amp.String()),
        ('password', amp.String()),
    ]
    response = [('result', amp.Unicode())]

class PythonLong(amp.Argument):
    """
    Convert to and from 'long'.
    """
    def toString(self, inObject):
        return pickle.encode_long(inObject)
    def fromString(self, inString):
        return pickle.decode_long(inString)

class GetPublicKey(amp.Command):
    '''
    Requests that the server send its public key for authentication. The public
    key values e and n are encoded using pickle.encode_long() before being sent
    as strings.
    '''
    arguments = []
    response = [
        ('e', PythonLong()),
        ('n', PythonLong()),
    ]



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.app import Application
from trosnoth.gui.framework import table
from trosnoth.gui.fonts.font import Font
from trosnoth.gui.common import *
from trosnoth.gui.framework import framework
import pygame

class Interface(framework.CompoundElement):
    def __init__(self, app):
        super(Interface, self).__init__(app)
        backdrop = pygame.Surface(app.screenManager.scaledSize)
        backdrop.fill((128,0,128))
        t1 = table.Table(app, Location(FullScreenAttachedPoint((-10,-10), 'bottomright'), 'bottomright'),
                         rows = 4, columns = 2)

        t1[0][0].setText("Hello")
        t1[0][0].style.font = Font('FreeSans.ttf', 16)
        t1[1][1].setText("There")
        t1[1][1].style.foreColour = (0,232, 0)
        t1[1][1].style.textAlign = 'center'
        t1[1][1].style.hasShadow = True
        t1[1][1].style.shadowColour = (0,0,255)
        t1[1][2].style.backColour = (232,0,0)
        t1[1][2].style.font = Font('FreeSans.ttf', 50)
        t1[1][2].setText("TOO BIG!!")
        t1[1][2].style.foreColour = (255,255,255)
        t1.getRow(1).setHeight(100)
        t1.getRow(3).setHeight(ScaledScalar(60))
        t1.setBorderWidth(7)
        self.elements = [t1]

size = (600,450)

if __name__ == '__main__':
    a = Application(size, 0, "Testing", Interface)
    a.run()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import pygame
import trosnoth.data.sound as sound
from trosnoth.data import getPath

class SoundAction(object):
    def __init__(self, filename, startChannelIndex, numChannels):
        assert numChannels > 0
        if not pygame.mixer.get_init():
            self.channels = []
            return

        self.sound = pygame.mixer.Sound(getPath(sound, filename))
        endChannelIndex = startChannelIndex+numChannels
        self.currentIndex = 0
        self.channels = [pygame.mixer.Channel(i) for i in
                range(startChannelIndex, endChannelIndex)]
        self.channelVolume = 1

    def play(self, volume=1):
        if not pygame.mixer.get_init():
            return

        finalVol = volume * self.channelVolume
        if finalVol < 0.01:
            return
        # Put the first channel to the back
        channel = self.channels[self.currentIndex]
        self.currentIndex += 1
        self.currentIndex %= len(self.channels)
        channel.set_volume(finalVol)
        channel.play(self.sound)

    def setVolume(self, val):
        self.sound.set_volume(val)

    def setChannelsVolume(self, val):
        self.channelVolume = val
        for channel in self.channels:
            channel.set_volume(val)

class SoundPlayer(object):
    def __init__(self):
        self.sounds = {}
        self.masterVolume = 1
        self._channelsUsed = 0
        pygame.mixer.set_num_channels(16)        

    def addSound(self, filename, action, maxSimultaneous = 8):
        if not pygame.mixer.get_init():
            return

        self.sounds[action] = SoundAction(filename, self._channelsUsed,
                maxSimultaneous)
        # In case a sound is added after the volume has been set:
        self.sounds[action].setChannelsVolume(self.masterVolume)

        self._channelsUsed += maxSimultaneous
        if self._channelsUsed > pygame.mixer.get_num_channels():
            pygame.mixer.set_num_channels(self._channelsUsed)

    def play(self, action, volume = 1):
        if not pygame.mixer.get_init():
            return

        self.sounds[action].play(volume)

    def setSoundVolume(self, action, val):
        self.sounds[action].setVolume(val)

    def setMasterVolume(self, val):
        self.masterVolume = val
        for action in self.sounds.values():
            action.setChannelsVolume(val)



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from trosnoth.gui.app import Application
from trosnoth.gui.framework import framework, prompt, hint, elements
from trosnoth.gui.fonts.font import Font
from trosnoth.gui.common import Area
import pygame
from random import randint, shuffle

class Interface(framework.CompoundElement):
    def __init__(self, app):
        super(Interface, self).__init__(app)
        backdrop = pygame.Surface(app.screenManager.scaledSize)
        backdrop.fill((128,0,128))
        i1 = prompt.InputBox(self.app, Area((40,40),(300,50)), 'Mouse Over', font = Font("KLEPTOCR.TTF", 36))
        i1.onClick.addListener(self.setFocus)
        i2 = prompt.InputBox(self.app, Area((340,140),(200,50)), 'Mouse Over', font = Font("KLEPTOCR.TTF", 36))
        i2.onClick.addListener(self.setFocus)
        h1 = hint.Hint(self.app, "You absolutely\nRULE!", i1, Font("KLEPTOCR.TTF", 24))
        h2 = hint.Hint(self.app, "Like, amazingly so!! TROSNOTH!@!!@!", i2, Font("KLEPTOCR.TTF", 24))
        h3 = hint.Hint(self.app, "A secret!\nWell done, you have found a magical land of wonder and amazement!! \nYou win absolutely nothing, but at least you have a good story to tell now.", pygame.Rect(500, 400, 100, 50), Font("KLEPTOCR.TTF", 24))
        self.elements = [elements.PictureElement(app, backdrop),
                         i1, h1,
                         i2, h2, h3
                         ]

size = (600,450)

if __name__ == '__main__':
    a = Application(size, 0, "Testing", Interface)
    a.run()



import logging

from twisted.internet import defer

from trosnoth import dbqueue
from trosnoth.messages import AchievementUnlockedMsg
from trosnoth.utils.twist import WeakLoopingCall
from trosnoth.utils.utils import timeNow

from trosnoth.gamerecording.achievementlist import availableAchievements

log = logging.getLogger(__name__)


class PlayerAchievements(object):
    '''Tracks achievements for a single player.'''
    achievementDefs = availableAchievements

    def __init__(self, manager, player):
        self.manager = manager
        self.player = player
        self.world = manager.world
        self.achievements = {
            c.idstring: c(player, self.world)
            for c in self.achievementDefs.perPlayer
        }
        self.stopped = self.started = False
        self.savingProgress = False

    def rejoined(self, player):
        self.player = player
        for achievement in self.achievements.itervalues():
            achievement.rejoined(player)

    def start(self):
        assert not self.started
        self.started = True
        for achievement in self.achievements.itervalues():
            achievement.onUnlocked.addListener(
                self.manager.achievementUnlocked)
            achievement.start()

    def stop(self):
        assert not self.stopped
        self.stopped = True
        for achievement in self.achievements.itervalues():
            achievement.stop()
            achievement.onUnlocked.removeListener(
                self.manager.achievementUnlocked)

    def saveProgress(self, force=False):
        if self.savingProgress and not force:
            return
        if self.player.user is None:
            return

        @dbqueue.add
        @defer.inlineCallbacks
        def saveAchievementProgressToDB():
            self.savingProgress = True
            try:
                for achievement in self.achievements.itervalues():
                    achievement.save()
                    yield dbqueue.wait()
            finally:
                self.savingProgress = False

    def __str__(self):
        return str(self.achievements.values)


class AchievementManager(object):
    SAVE_PERIOD = 20

    achievementDefs = availableAchievements

    def __init__(self, game):
        self.game = game
        self.world = world = game.world
        self.lastSave = timeNow()

        # All players recorded this game, indexed by user id.
        self.allPlayers = {}

        defs = self.achievementDefs
        self.oncePerGameAchievements = [c(world) for c in defs.oncePerGame]
        self.oncePerTeamPerGameAchievements = [
            c(world, team)
            for c in defs.oncePerTeamPerGame
            for team in world.teams
        ]

        self.loop = WeakLoopingCall(self, 'save')
        self.started = False
        self.stopped = False

    def save(self, force=False):
        for playerAchievements in self.allPlayers.itervalues():
            playerAchievements.saveProgress(force=force)

    def start(self):
        assert not self.started
        self.started = True
        for player in self.world.players:
            self.playerAdded(player)

        self.world.onPlayerAdded.addListener(self.playerAdded)
        self.loop.start(self.SAVE_PERIOD)

        for a in self.oncePerGameAchievements:
            a.onUnlocked.addListener(self.achievementUnlocked)
            a.start()
        for a in self.oncePerTeamPerGameAchievements:
            a.start()
            a.onUnlocked.addListener(self.achievementUnlocked)

    def stop(self):
        assert not self.stopped
        self.stopped = True
        self.loop.stop()
        self.world.onPlayerAdded.removeListener(self.playerAdded)
        self.save(force=True)
        for a in self.oncePerGameAchievements:
            a.stop()
            a.onUnlocked.removeListener(self.achievementUnlocked)
        for a in self.oncePerTeamPerGameAchievements:
            a.stop()
            a.onUnlocked.removeListener(self.achievementUnlocked)
        for p in self.allPlayers.itervalues():
            p.stop()

    def playerAdded(self, player):
        name = player.identifyingName
        if name not in self.allPlayers:
            a = PlayerAchievements(self, player)
            self.allPlayers[name] = a
            a.start()
        else:
            self.allPlayers[name].rejoined(player)

    def triggerAchievement(self, player, achievementId):
        '''
        Called from stats manager to trigger a stats-related achievement.
        '''
        a = self.allPlayers[player.identifyingName].achievements[achievementId]
        a.achievementTriggered()

    def achievementUnlocked(self, achievement, player):
        if player.user is not None:
            player.user.achievementUnlocked(achievement.idstring)
        self.game.sendServerCommand(
            AchievementUnlockedMsg(player.id, achievement.idstring.encode()))



import json
import logging

from trosnoth.gamerecording import replays, stats, statgeneration
from trosnoth.messages import SetTeamNameMsg
from trosnoth.network.networkDefines import validServerVersions, serverVersion
from trosnoth.data import getPath, user, makeDirs
import os
import time

log = logging.getLogger(__name__)

gameDir = 'recordedGames'
gameExt = '.tros'
replayDir = os.path.join(gameDir, 'replays')
replayExt = '.trosrepl'
statDir = os.path.join(gameDir, 'stats')
statExt = '.trosstat'
htmlDir = os.path.join(gameDir, 'htmlStats')


def getFilename(alias, directory, ext, multipleFiles = True):
    # Figure out the filename to use for the main file
    gamePath = getPath(user, directory)
    makeDirs(gamePath)
    copyCount = 0
    succeeded = False
    if multipleFiles:
        while not succeeded:
            filename = '%s (%s)%s' % (alias, str(copyCount), ext)
            filePath = os.path.join(gamePath, filename)
            succeeded = not os.path.exists(filePath)
            copyCount += 1
    else:
        filename = '%s%s' % (alias, ext)
        filePath = os.path.join(gamePath, filename)
    return filePath

# Separate from the server version, which is to do with the
# types and content of network messages.
recordedGameVersion = 2


class RecordedGameException(Exception):
    pass


class GameRecorder(object):

    def __init__(
            self, game, world, saveReplay=False, gamePrefix=None,
            replayPath=replayDir):
        super(GameRecorder, self).__init__()
        if gamePrefix is None:
            gamePrefix = 'unnamed'
        self.alias = gamePrefix + ' game'
        self.game = game
        self.serverVersion = serverVersion
        self.world = world
        self.saveReplay = saveReplay
        if replayPath is None:
            replayPath = replayDir
        self.replayPath = replayPath
        self.currentlySaving = False

    def consumeMsg(self, msg):
        if self.currentlySaving:
            self.replayRecorder.consumeMsg(msg)
            if isinstance(msg, SetTeamNameMsg):
                self.gameFile.teamNameChanged(msg.teamId, msg.name)

    def gotStateChange(self):
        if (
                self.world.state.startsNewMatch
                or self.world.state.inMatch != self.currentlySaving):
            self.saveEverything()
            self.stop()
            self.start()

    def saveEverything(self):
        if self.currentlySaving:
            self.statRecorder.save()

    def reset(self):
        saveReplay = self.saveReplay
        if not self.world.state.inMatch:
            saveReplay = False
        self.currentlySaving = saveReplay

        self.filename = getFilename(self.alias, gameDir, gameExt)
        self.replayFilename = getFilename(
            self.alias, self.replayPath, replayExt)
        self.statsFilename = getFilename(self.alias, statDir, statExt)
        if saveReplay:
            self.gameFile = RecordedGameFile(self.filename, sVersion=serverVersion,
                    alias=self.alias, replayFilename=self.replayFilename,
                    statsFilename=self.statsFilename,
                    halfMapWidth=self.world.map.layout.halfMapWidth,
                    mapHeight=self.world.map.layout.mapHeight)
            self.gameFile.save()
            self.statRecorder = stats.StatKeeper(
                self, self.world, self.statsFilename)
            self.replayRecorder = replays.ReplayRecorder(self.world,
                    self.replayFilename)

    def start(self):
        self.world.onGameStateChanged.addListener(self.gotStateChange)
        self.reset()

    def stop(self):
        self.world.onGameStateChanged.removeListener(self.gotStateChange)
        if self.currentlySaving:
            self.replayRecorder.stop()
            self.statRecorder.stop()
            self.gameFile.gameFinished()


class RecordedGame(object):
    def __init__(self, filename):
        self.filename = filename
        self.gameFile = RecordedGameFile(self.filename)
        self.gameFile.load()
        if not self.gameFile.isValid():
            raise RecordedGameException('Invalid Game File')

    def serverInformation(self):
        return self.gameFile.serverInformation

    def wasFinished(self):
        return self.gameFile.wasFinished()

    def generateHtmlFile(self):
        '''
        Generate an html stats file, and return the url
        '''
        if self.gameFile.hasHtml():
            return self.gameFile.htmlFile
        self.htmlFile = getFilename(self.gameFile.alias, htmlDir, '.html')
        statgeneration.generateHtml(self.htmlFile, self.gameFile.statsFilename)
        self.gameFile.htmlGenerated(self.htmlFile)
        return self.htmlFile

    def __getattr__(self, attr):
        return getattr(self.gameFile, attr)


class RecordedGameFile(object):
    def __init__(self, filename, sVersion=None, alias=None, replayFilename=None,
            statsFilename=None, halfMapWidth=None, mapHeight=None):
        self.filename = filename
        self.serverInformation = {}
        self.serverInformation['recordedGameVersion'] = recordedGameVersion
        self.serverInformation['serverVersion'] = sVersion
        self.serverInformation['alias'] = alias
        self.serverInformation['replayFilename'] = replayFilename
        self.serverInformation['statsFilename'] = statsFilename
        self.serverInformation['dateTime'] = ','.join(map(str,time.localtime()))
        self.serverInformation['unixTimestamp'] = time.time()
        self.serverInformation['halfMapWidth'] = halfMapWidth
        self.serverInformation['mapHeight'] = mapHeight
        self.serverInformation['teamAname'] = 'Blue'
        self.serverInformation['teamBname'] = 'Red'

    def gameFinished(self):
        self.serverInformation['gameFinishedTimestamp'] = time.time()
        self.save()

    def teamNameChanged(self, teamId, teamName):
        self.serverInformation['team%sname' % (teamId,)] = teamName
        self.save()

    def htmlGenerated(self, filename):
        self.serverInformation['htmlFile'] = filename
        self.save()

    ##
    # Overwrites any existing file
    def save(self):
        # No value may be null
        for value in self.serverInformation.itervalues():
            assert value is not None
        with open(self.filename, 'w') as f:
            serverInfoString = json.dumps(self.serverInformation, indent=4)
            f.write(serverInfoString)

    def load(self):
        file = open(self.filename, 'r')
        lines = file.readlines()

        fullText = '\n'.join(lines)
        self.serverInformation = json.loads(fullText)

    def isValid(self):
        return self.serverInformation['serverVersion'] in validServerVersions

    def wasFinished(self):
        return self.serverInformation.has_key('gameFinishedTimestamp')

    def hasHtml(self):
        return self.serverInformation.has_key('htmlFile')

    def __getattr__(self, key):
        return self.serverInformation[key]



from trosnoth.gui.common import Location, Area, ScaledPoint, ScaledSize
from trosnoth.gui.notify import NotificationBar


class FirstPlayNotificationBar(NotificationBar):
    '''
    Provides a message for first-time players
    '''
    def __init__(self, app):
        super(FirstPlayNotificationBar, self).__init__(app,
            message =
                'First time playing Trosnoth? Click here to learn the rules.',
            url = 'http://www.trosnoth.org/how-to-play',
            font = app.fonts.default,
            area = Area(
                ScaledPoint(0, 0),
                ScaledSize(1024, 30),
                'topleft'
            ),
            buttonPos = Location(ScaledPoint(1024, 0), 'topright'),
            textPos = Area(
                ScaledPoint(512, 15),
                ScaledSize(1024, 30),
                'centre'
            ),
        )
        self.onClick.addListener(self.hide)
        self.onClose.addListener(app.identitySettings.notFirstTime)



import logging
import sys

import pygame
from twisted.internet import defer, reactor

from trosnoth.game import LocalGame, RemoteGame
from trosnoth.gui import app
from trosnoth.settings import DisplaySettings, SoundSettings, IdentitySettings
from trosnoth.themes import Theme
from trosnoth.trosnothgui.ingame.gameInterface import GameInterface
from trosnoth.gui.framework import framework
from trosnoth.manholehelper import LocalManholeHelper
from trosnoth.model import mapLayout
from trosnoth.model.agenthub import LocalHub
from trosnoth.model.hub import Hub, Node
from trosnoth.model.uithrottler import UIMsgThrottler, LocalGameTweener

log = logging.getLogger(__name__)

GAME_TITLE = 'Trosnoth Solo Mode'


class Main(app.MultiWindowApplication):
    def __init__(
            self, isolate=False, size=(1, 1), aiCount=0, aiClass='alpha',
            mapBlocks=(), testMode=False, lobby=False, stackTeams=False,
            trosball=False, delay=None, blockRatio=0.5, duration=0):
        '''Initialise the game.'''
        pygame.init()
        self.size = size
        self.aiCount = aiCount
        self.isolate = isolate
        self.mapBlocks = mapBlocks
        self.testMode = testMode
        self.lobby = lobby
        self.stackTeams = stackTeams
        self.aiClass = aiClass
        self.delay = delay
        self.blockRatio = blockRatio
        self.duration = duration

        self.displaySettings = DisplaySettings(self)
        self.soundSettings = SoundSettings(self)
        self.identitySettings = IdentitySettings(self)
        self.ais = []

        pygame.font.init()
        super(Main, self).__init__(
            self.displaySettings.getSize(),
            self.displaySettings.fullScreen,
            GAME_TITLE,
            SoloInterface)

        # Set the master sound volume.
        self.soundSettings.apply()

        self.game = None
        self.startGame(trosball)

    def getConsoleLocals(self):
        result = {
            'game': self.game,
            'ais': self.ais,
            'helper': LocalManholeHelper(lambda: self.game),
        }
        return result

    def initialise(self):
        super(Main, self).initialise()

        # Loading the theme loads the fonts.
        self.theme = Theme(self)

    def getFontFilename(self, fontName):
        '''
        Tells the UI framework where to find the given font.
        '''
        return self.theme.getPath('fonts', fontName)

    def changeScreenSize(self, size, fullScreen):
        self.screenManager.setScreenProperties(size, fullScreen, GAME_TITLE)

    def startGame(self, trosball=False):
        db = mapLayout.LayoutDatabase(
            pathGetter=self.theme.getPath,
            blocks=self.mapBlocks)
        self.game = game = LocalGame(
            db, self.size[0], self.size[1], onceOnly=True,
            blockRatio=self.blockRatio,
            duration=self.duration * 60,
        )
        if trosball:
            self.game.world.setNextGameType('solotrosball')
        else:
            self.game.world.setNextGameType('solo')
        if not self.lobby:
            game.world.beginGameStartCountdown()
        if self.testMode:
            game.world.setTestMode()

        # Create a client and an interface.
        if self.isolate:
            rgame = RemoteGame(db)
            hub = LocalHub(game)
            self.tweener = UIMsgThrottler()
            if self.delay:
                delayer = DelayNodeHub(self.delay)
                hub.connectNode(delayer)
                delayer.connectNode(self.tweener)
            else:
                hub.connectNode(self.tweener)
            self.tweener.connectNode(rgame)
            self.rgame = rgame
            self.gi = gi = GameInterface(self, rgame)
            rgame.connected(game.world.dumpEverything())
        else:
            self.tweener = LocalGameTweener(game)
            self.gi = gi = GameInterface(self, game)
        gi.onDisconnectRequest.addListener(self.stop)
        gi.onConnectionLost.addListener(self.stop)
        self.interface.elements.append(gi)

        self.ais[:] = []

        try:
            for i in xrange(self.aiCount):
                if self.stackTeams:
                    ai = game.addAI(self.aiClass, team=0)
                else:
                    ai = game.addAI(self.aiClass)
                self.ais.append(ai)
        except ImportError:
            print >> sys.stderr, 'AI module not found: %s' % (self.aiClass,)
            sys.exit(1)
        except AttributeError:
            print >> sys.stderr, (
                'AI module does not define AIClass: %s' % (self.aiClass,))
            sys.exit(1)

    def stop(self):
        self.game.stop()
        super(Main, self).stop()


class DelayNodeHub(Hub, Node):
    def __init__(self, delay, *args, **kwargs):
        super(DelayNodeHub, self).__init__(*args, **kwargs)
        self.delay = delay

    @defer.inlineCallbacks
    def connectNewAgent(self):
        result = yield self.hub.connectNewAgent()

        d = defer.Deferred()
        reactor.callLater(self.delay, d.callback, None)
        yield d

        defer.returnValue(result)

    def disconnectAgent(self, agentId):
        reactor.callLater(self.delay, self.hub.disconnectAgent, agentId)

    def sendRequestToGame(self, agentId, msg):
        msg.tracePoint(self, 'sendRequestToGame')
        reactor.callLater(self.delay, self.hub.sendRequestToGame, agentId, msg)

    def gotServerCommand(self, msg):
        msg.tracePoint(self, 'gotServerCommand')
        reactor.callLater(self.delay, self.node.gotServerCommand, msg)

    def gotMessageToAgent(self, agentId, msg):
        msg.tracePoint(self, 'gotMessageToAgent')
        reactor.callLater(
            self.delay, self.node.gotMessageToAgent, agentId, msg)

    def agentDisconnected(self, agentId):
        reactor.callLater(self.delay, self.node.agentDisconnected, agentId)


class SoloInterface(framework.CompoundElement):
    def __init__(self, app):
        super(SoloInterface, self).__init__(app)
        self.elements = []



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import logging

import pygame

from trosnoth.gui.framework.framework import Element, CompoundElement
from trosnoth.gui.common import Location, AttachedPoint

log = logging.getLogger('scrollableCanvas')

class ScrollBar(Element):
    defaultWidth = 15
    defaultButtonChange = 20  # number of pixels moved when button is pressed
    defaultColour = (0, 192, 0)
    defaultBackColour = (128, 128, 128)
    def __init__(self, app, parent, pos, horizontal=False):
        Element.__init__(self, app)
        self.width = self.defaultWidth
        self.buttonChange = self.defaultButtonChange
        self.colour = self.defaultColour
        self.backColour = self.defaultBackColour
        self.parent = parent
        self.pos = pos
        self.horizontal = horizontal
        if horizontal:
            self.index = 0
        else:
            self.index = 1

        self.originalPosition = 0
        self.position = 0
        self.beingDraggedFrom = None

    def _getWidth(self):
        return self.width

    # This should be checked before ever being used
    def __checkThisIsValid(self):
        if self.__getDrawableSize() >= self.__getCanvasSize():
            log.error(("Vertical", "Horizontal")[self.horizontal])
            log.error("Drawable Area: %d", self.__getDrawableSize())
            log.error("Canvas Size: %d", self.__getCanvasSize())
            assert False, ('For there to be a scroll bar, the displaying '
                    'size must be less than the full size')

    def _getTopLeftPt(self):
        self.__checkThisIsValid()
        return self._getFullRect().topleft

    def _getBarLength(self):
        self.__checkThisIsValid()
        return self.__getDrawableSize() ** 2 / self.__getCanvasSize()

    def _getBarRect(self):
        self.__checkThisIsValid()
        if self.horizontal:
            size = (self._getBarLength(), self._getWidth())
            pos = (self._getTopLeftPt()[0] + self.getPosition(),
                    self._getTopLeftPt()[1])
        else:
            size = (self._getWidth(), self._getBarLength())
            pos = (self._getTopLeftPt()[0], self._getTopLeftPt()[1] +
                    self.getPosition())
        return pygame.rect.Rect(pos, size)

    def __getDrawableSize(self):
        return self.parent._getDrawableSize()[self.index]

    def __getCanvasSize(self):
        return self.parent._getCanvasSize()[self.index]

    def _getSize(self):
        self.__checkThisIsValid()
        if self.horizontal:
            x = self.__getDrawableSize()
            y = self._getWidth()
        else:
            x = self._getWidth()
            y = self.__getDrawableSize()
        return (x,y)

    def _getFullRect(self):
        self.__checkThisIsValid()
        r = pygame.rect.Rect((0,0), self._getSize())
        self.pos.apply(self.app, r)
        return r

    def _movedTo(self, pos):
        self.__checkThisIsValid()
        if self.horizontal:
            self.setPosition(self.originalPosition + (pos[0] -
                    self.beingDraggedFrom[0]))
        else:
            self.setPosition(self.originalPosition + (pos[1] -
                    self.beingDraggedFrom[1]))

        newScrollingPos = (self.getPosition() * self.__getCanvasSize() /
                self.__getDrawableSize())
        newScrollingPos = min(newScrollingPos, self.__getCanvasSize() -
                self.__getDrawableSize())
        if self.horizontal:
            self.parent.setHorizontalPos(newScrollingPos)
        else:
            self.parent.setVerticalPos(newScrollingPos)

    def getPosition(self):
        # This line is useful in the event of a screen resize
        self.setPosition(self.position)
        return self.position

    def setPosition(self, pos):
        self.position = min(max(pos, 0), self.__getDrawableSize() -
                self._getBarLength())

    def processEvent(self, event):
        self.__checkThisIsValid()
        if (event.type == pygame.MOUSEBUTTONDOWN and event.button == 1
                and self._getBarRect().collidepoint(event.pos) and
                self.beingDraggedFrom == None):
            self.beingDraggedFrom = event.pos
            return None
        elif (event.type == pygame.MOUSEMOTION and self.beingDraggedFrom
                != None):
            self._movedTo(event.pos)
            return None
        elif (event.type == pygame.MOUSEBUTTONUP and event.button == 1
                and self.beingDraggedFrom != None):
            self.beingDraggedFrom = None
            self.originalPosition = self.getPosition()
            return None
        return event

    def draw(self, screen):
        self.__checkThisIsValid()
        pygame.draw.rect(screen, self.backColour, self._getFullRect(), 0)
        pygame.draw.rect(screen, self.colour, self._getBarRect(), 0)


class ScrollableCanvas(CompoundElement):
    '''
    For a scrollable canvas, the displaySize is how big it will actually be
    drawn, while its size is how big the canvas is in memory.  If the size is
    bigger than the displaySize either horizontally or vertically, there will be
    a scroll bar to allow moving around the full area.
    '''
    def __init__(self, app, pos, size, displaySize):
        CompoundElement.__init__(self, app)

        self.pos = pos
        self.size = size
        self.displaySize = displaySize
        self.backgroundColour = (64,64,64)

        self._screenSize = tuple(app.screenManager.size)

        self.__horizontalScrollBar = ScrollBar(app, self,
                Location(AttachedPoint((0,0), self._getRect, 'bottomleft'),
                'bottomleft'), True)
        self.__verticalScrollBar = ScrollBar(app, self,
                Location(AttachedPoint((0,0), self._getRect, 'topright'),
                'topright'), False)

        # internalPos is the position on the full canvas that we are drawing
        # from
        self.internalPos = None
        self.setInternalPos((0,0))

    def getInternalPos(self):
        if self._screenSize != self.app.screenManager.size:
            self._screenSize = tuple(self.app.screenManager.size)
            self.setInternalPos(self.internalPos)

        return self.internalPos

    def setInternalPos(self, pos):
        self.internalPos = tuple([min(pos[i], (self._getCanvasSize()[i] -
                self._getDrawableSize()[i])) for i in 0,1])

    def _getHorizontalScrollBar(self):
        if self._needsHorizontalScrollBar():
            return self.__horizontalScrollBar
        return None

    def _getVerticalScrollBar(self):
        if self._needsVerticalScrollBar():
            return self.__verticalScrollBar
        return None

    def _calculateIfWeNeedBars(self):
        hScroll = self._getMinimumCanvasSize()[0] > self._getSize()[0]
        if hScroll:
            vScroll = (self._getMinimumCanvasSize()[1] > self._getSize()[1] -
                    self.__horizontalScrollBar._getWidth())
        else:
            vScroll = self._getMinimumCanvasSize()[1] > self._getSize()[1]
        if vScroll and not hScroll:
            # re-calculate now that we know we have a vertical scroll bar
            hScroll = (self._getMinimumCanvasSize()[0] > self._getSize()[0] -
                    self.__verticalScrollBar._getWidth())
        return (hScroll, vScroll)

    def _getDrawableSize(self):
        drawableHeight = self._getSize()[1]

        hScroll, vScroll = self._calculateIfWeNeedBars()

        if hScroll:
            drawableHeight -= ScrollBar.defaultWidth

        drawableWidth = self._getSize()[0]
        if vScroll:
            drawableWidth -= ScrollBar.defaultWidth

        return (drawableWidth, drawableHeight)

    def _needsHorizontalScrollBar(self):
        return self._calculateIfWeNeedBars()[0]

    def _needsVerticalScrollBar(self):
        return self._calculateIfWeNeedBars()[1]

    def _getPt(self):
        return self._getRect().topleft

    def _getSize(self):
        return self.displaySize.getSize(self.app)

    def _getMinimumCanvasSize(self):
        return self.size.getSize(self.app)

    def _getCanvasSize(self):
        canvasSize = self._getMinimumCanvasSize()
        availableSize = self._getDrawableSize()
        x = max(availableSize[0], canvasSize[0])
        y = max(availableSize[1], canvasSize[1])
        return x, y

    def _getRect(self):
        r = pygame.rect.Rect((0,0), self._getSize())
        self.pos.apply(self.app, r)
        return r

    def setVerticalPos(self, verticalPos):
        self.setInternalPos((self.getInternalPos()[0], verticalPos))

    def setHorizontalPos(self, horizontalPos):
        self.setInternalPos((horizontalPos, self.getInternalPos()[1]))

    def _getContentsRect(self):
        return pygame.rect.Rect(self._getPt(), self._getDrawableSize())

    def _getOffsetCanvasRect(self):
        topleft = self._getPt()
        offsetPt = tuple([topleft[i] - self.getInternalPos()[i] for i in 0, 1])
        return pygame.rect.Rect(offsetPt, self._getCanvasSize())

    def draw(self, screen):
        contentsRect = self._getContentsRect()
        screen.fill(self.backgroundColour, contentsRect)
        if self._needsHorizontalScrollBar():
            self._getHorizontalScrollBar().draw(screen)
        if self._needsVerticalScrollBar():
            self._getVerticalScrollBar().draw(screen)
        oldClipRect = screen.get_clip()

        if oldClipRect is None:
            screen.set_clip(contentsRect)
        else:
            assert isinstance(oldClipRect, pygame.rect.Rect)
            r = oldClipRect.clip(contentsRect)
            screen.set_clip(r)
        try:
            CompoundElement.draw(self, screen)
        finally:
            screen.set_clip(oldClipRect)

    def processEvent(self, event):
        if self._needsHorizontalScrollBar():
            event = self._getHorizontalScrollBar().processEvent(event)
        if self._needsVerticalScrollBar() and event:
            event = self._getVerticalScrollBar().processEvent(event)
        if event:
            if (event.type == pygame.MOUSEBUTTONDOWN or event.type ==
                    pygame.MOUSEBUTTONUP):
                if self._getContentsRect().collidepoint(event.pos):
                    CompoundElement.processEvent(self, event)
                    return None
                elif event.type == pygame.MOUSEMOTION:
                    CompoundElement.processEvent(self, event)
                    return event
                else:
                    return event
            else:
                return CompoundElement.processEvent(self, event)
        else:
            return None

def ScrollableCanvasAttachedPoint(scrollableCanvas, val, anchor='topleft'):
    return AttachedPoint(val, scrollableCanvas._getOffsetCanvasRect, anchor)



from trosnoth.gui.framework import framework
from trosnoth.trosnothgui.common import setAlpha
import pygame


class GaugeBase(framework.Element):
    def __init__(self, app, area):
        super(GaugeBase, self).__init__(app)
        self.area = area

        self.icon = self.getIcon()
        if self.icon is not None:
            self.icon = self.icon.copy()
            setAlpha(self.icon, 160)

    def draw(self, surface):
        rect = self.area.getRect(self.app)
        pos = rect.topleft
        ratio = min(1, max(0, self.getRatio()))
        amount = int(ratio * rect.width)

        backColour = self.getBackColour()
        if backColour is not None:
            backRect = pygame.rect.Rect(
                pos[0] + amount, pos[1], rect.width - amount + 1, rect.height)
            surface.fill(backColour, backRect)

        if amount > 0:
            insideRect = pygame.rect.Rect(pos, (amount, rect.height))
            surface.fill(self.getForeColour(), insideRect)

        # Draw the border on top
        pygame.draw.rect(surface, self.app.theme.colours.gaugeBorder, rect, 2)

        if self.icon is not None:
            r = self.icon.get_rect()
            r.center = rect.midleft
            r.left -= r.width // 5
            surface.blit(self.icon, r)

    def getRatio(self):
        '''
        Return a number as a proportion (0..1) of how complete
        this box is. To be implemented in subclasses
        '''
        raise NotImplementedError

    def getForeColour(self):
        '''
        Return the foreground colour that this gauge should be.
        To be implemented in subclasses
        '''
        raise NotImplementedError

    def getBackColour(self):
        '''
        Return the background colour that this gauge should be.
        None = blank
        To be implemented in subclasses
        '''
        return None

    def getIcon(self):
        return None


class RespawnGauge(GaugeBase):
    '''Represents a graphical gauge to show how close to respawning a player
    is.'''
    def __init__(self, app, area, player, world):
        self.player = player
        self.world = world
        super(RespawnGauge, self).__init__(app, area)

    def getRatio(self):
        if self.world.state.isStartingCountdown():
            value = self.world.timer.getStartCountdown()
            max = self.world.timer.getStartCountdownMax()
            return 1 - value / max
        else:
            return 1 - (
                self.player.respawnGauge /
                self.player.world.physics.playerRespawnTotal)

    def getForeColour(self):
        if self.getRatio() >= 1:
            return self.app.theme.colours.gaugeGood
        else:
            return self.app.theme.colours.gaugeBad

    def getIcon(self):
        return self.app.theme.sprites.ghostIcon(self.player.team).getImage()


class TurretGauge(GaugeBase):
    '''Represents a graphical gauge to show the overheatedness of a turret'''
    def __init__(self, app, area, player):
        self.player = player
        super(TurretGauge, self).__init__(app, area)

    def getRatio(self):
        return (self.player.turretHeat /
                self.player.world.physics.playerTurretHeatCapacity)

    def getForeColour(self):
        if self.player.turretOverHeated:
            return self.app.theme.colours.gaugeBad
        elif self.getRatio() > 0.5:
            return self.app.theme.colours.gaugeWarn
        else:
            return self.app.theme.colours.gaugeGood

    def getIcon(self):
        return self.app.theme.sprites.turretBase.getImage()


class GunGauge(GaugeBase):
    player = None

    def getRatio(self):
        player = self.player
        if player is None:
            return 0

        if player.machineGunner and player.mgBulletsRemaining > 0:
            return player.mgBulletsRemaining / 15.

        if player.reloadTime > 0:
            return 1 - player.reloadTime / (player.reloadFrom + 0.)

        return 1

    def getForeColour(self):
        player = self.player
        if player.machineGunner and player.mgBulletsRemaining > 0:
            return self.app.theme.colours.gaugeGood
        if player is None or player.reloadTime > 0:
            return self.app.theme.colours.gaugeBad
        return self.app.theme.colours.gaugeGood

    def getIcon(self):
        return self.app.theme.sprites.gunIcon.getImage()


class UpgradeGauge(GaugeBase):
    '''Represents a graphical gauge to show how much time a player has left
    to use their upgrade.'''
    def __init__(self, app, area, player):
        self.player = player
        super(UpgradeGauge, self).__init__(app, area)

    def getRatio(self):
        if self.player.upgradeTotal == 0:
            return 1
        return self.player.upgradeGauge / self.player.upgradeTotal

    def getForeColour(self):
        return self.app.theme.colours.gaugeGood

    def getIcon(self):
        return self.app.theme.sprites.upgradeImage(type(self.player.upgrade))


class StarGauge(GaugeBase):
    '''
    Shows how your teams stars are going compared to the number required for
    the selected upgrade.
    '''

    def __init__(self, app, area, player, upgrade):
        self.player = player
        self.upgrade = upgrade
        super(StarGauge, self).__init__(app, area)

    def getRatio(self):
        if not self.upgrade.enabled:
            return 1
        if self.upgrade.requiredStars == 0:
            return 1
        return self.player.getTeamStars() / (self.upgrade.requiredStars + 0.)

    def getForeColour(self):
        if not self.upgrade.enabled:
            return self.app.theme.colours.gaugeBad
        if self.getRatio() < 1:
            return self.app.theme.colours.gaugeBad
        return self.app.theme.colours.gaugeGood
